# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
import wfdb

from ...cfg import CFG
from ...utils.misc import add_docstring
from ...utils.utils_data import ECGWaveForm
from ..base import (
    PhysioNetDataBase,
    DataBaseInfo,
    BeatAnn,
    WFDB_Beat_Annotations,
    WFDB_Non_Beat_Annotations,
)


__all__ = [
    "QTDB",
]


_QTDB_INFO = DataBaseInfo(
    title="""
    QT Database
    """,
    about="""
    1. The QT Database includes ECGs which were chosen to represent a wide variety of QRS and ST-T morphologies
    2. Recordings were chosen chosen from the MIT-BIH Arrhythmia Database (MITDB), the European Society of Cardiology ST-T Database (EDB), and several other ECG databases collected at Boston's Beth Israel Deaconess Medical Center (MIT-BIH ST Change Database, MIT-BIH Supraventricular Arrhythmia Database, MIT-BIH Normal Sinus Rhythm Database, MIT-BIH Long-Term ECG Database, "sudden death" patients from BIH)
    3. Contains 105 fifteen-minute two-lead ECG recordings
    4. Contains onset, peak, and end markers for P, QRS, T, and (where present) U waves of from 30 to 50 selected beats in each recording
    5. Annotation file table:

        +--------+------------------------------------------------------+
        | Suffix | Meaning                                              |
        +========+======================================================+
        | .atr   | reference beat annotations from original database    |
        |        | (not available for the 24 sudden death records)      |
        +--------+------------------------------------------------------+
        | .man:  | reference beat annotations for selected beats only   |
        +--------+------------------------------------------------------+
        | .q1c:  | manually determined waveform boundary measurements   |
        |        | for selected beats (annotator 1 only -- second pass) |
        +--------+------------------------------------------------------+
        | .q2c:  | manually determined waveform boundary measurements   |
        |        | for selected beats (annotator 2 only -- second pass; |
        |        | available for only 11 records)                       |
        +--------+------------------------------------------------------+
        | .qt1:  | manually determined waveform boundary measurements   |
        |        | for selected beats (annotator 1 only -- first pass)  |
        +--------+------------------------------------------------------+
        | .qt2:  | manually determined waveform boundary measurements   |
        |        | for selected beats (annotator 2 only -- first pass;  |
        |        | available for only 11 records)                       |
        +--------+------------------------------------------------------+
        | .pu:   | automatically determined waveform boundary           |
        |        | measurements for all beats (based on both signals)   |
        +--------+------------------------------------------------------+
        | .pu0:  | automatically determined waveform boundary           |
        |        | measurements for all beats (based on signal 0 only)  |
        +--------+------------------------------------------------------+
        | .pu1:  | automatically determined waveform boundary           |
        |        | measurements for all beats (based on signal 1 only)  |
        +--------+------------------------------------------------------+

    6. A part of the recordings have rhythm annotations, ST change (elevation or depression) annotations, all of which have .atr annotation files. These annotations are provided in the `aux_note` attribute of the annotation object.
    7. In the first pass manual wave delineation annotation files (.qt1, .qt2 files), fiducial points were marked by a "|" symbol, along with beat annotations (one of "A", "B", "N", "Q", "V") inherited from corresponding .man files.
    8. In the second pass manual wave delineation annotation files (.q1c, .q2c files), the final manual annotations are recorded, with the regular annotation symbols "(" ,")", "t", "p", and "u", and with annotations inherited from the .qt1, .qt2 files.
    9. The .pu0, .pu1 files contain the automatic waveform onsets and ends in signals 0 and 1 respectively, as detected using the differentiated threshold method by ecgpuwave. In the num fields of the pu* annotations, ecgpuwave classifies the T waves as normal (0), inverted (1), only upwards (2), only downwards (3), biphasic negative-positive (4), or biphasic positive-negative (5). Waveform onset (and offset) annotations specify the waveform type in their num fields (0 for a P-wave, 1 for a QRS complex, 2 for a T wave, or 3 for a U-wave).
    10. Webpage of the database on PhysioNet [1]_. Paper describing the database [2]_.
    """,
    usage=[
        "ECG wave delineation",
        "ST segment",
    ],
    references=[
        "https://physionet.org/content/qtdb/1.0.0/",
        "Laguna P, Mark RG, Goldberger AL, Moody GB. A Database for Evaluation of Algorithms for Measurement of QT and Other Waveform Intervals in the ECG. Computers in Cardiology 24:673-676 (1997).",
    ],
    issues="""
    1. According to the paper of the database, there should be .ari files containing QRS annotations obtained automatically by ARISTOTLE, which however are not available in the database.
    2. A large proportion of the wave delineation annotations lack onset indices (the T waves and U waves).
    """,
    doi=[
        "10.1109/cic.1997.648140",
        "10.13026/C24K53",
    ],
)


@add_docstring(_QTDB_INFO.format_database_docstring(), mode="prepend")
class QTDB(PhysioNetDataBase):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "QTDB"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="qtdb",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.fs = 250
        self.spacing = 1000 / self.fs
        self.data_ext = "dat"

        # fmt: off
        self.all_extensions = ["atr", "man", "q1c", "q2c", "qt1", "qt2", "pu", "pu0", "pu1"]
        """
        1. .atr:    reference beat annotations from original database (not available in all cases)
        2. .man:    reference beat annotations for selected beats only
        3. .q1c:    manually determined waveform boundary measurements for selected beats (annotator 1 only -- second pass)
        4. .q2c:    manually determined waveform boundary measurements for selected beats (annotator 2 only -- second pass; available for only 11 records)
        5. .q1t:    manually determined waveform boundary measurements for selected beats (annotator 1 only -- first pass)
        6. .q2t:    manually determined waveform boundary measurements for selected beats (annotator 2 only -- first pass; available for only 11 records)
        7. .pu:     automatically determined waveform boundary measurements for all beats (based on both signals)
        8. .pu0:    automatically determined waveform boundary measurements for all beats (based on signal 0 only)
        9. .pu1:    automatically determined waveform boundary measurements for all beats (based on signal 1 only)
        """

        # records have different lead names
        # therefore, self.all_leads should not be set
        # otherwise, it will cause problems when loading data using `self.load_data`
        self._all_leads = [
            "CC5", "CM2", "CM4", "CM5", "D3", "D4", "ECG1", "ECG2", "ML5", "MLII",
            "V1", "V1-V2", "V2", "V2-V3", "V3", "V4", "V4-V5", "V5", "mod.V1",
        ]
        self.all_annotations = ["(", ")", "N", "t", "p"]
        # fmt: on

        self.beat_types_extended = list("""~"+/AFJNQRSTVaefjs|""")
        self.nonbeat_types = [
            item
            for item in self.beat_types_extended
            if item in WFDB_Non_Beat_Annotations
        ]
        self.beat_types = [
            item for item in self.beat_types_extended if item in WFDB_Beat_Annotations
        ]
        self.beat_types_map = {item: i for i, item in enumerate(self.beat_types)}
        self.beat_types_extended_map = {
            item: i for i, item in enumerate(self.beat_types_extended)
        }

        self.class_map = CFG(p=1, N=2, t=3, i=0)  # an extra isoelectric

        self._ls_rec()

    def get_subject_id(self, rec: Union[str, int]) -> int:
        """Attach a unique subject ID for the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        int
            Subject ID associated with the record.

        """
        raise NotImplementedError

    def get_lead_names(self, rec: Union[str, int]) -> List[str]:
        """Get the lead names of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        List[str]
            List of the lead names of the record.

        """
        return wfdb.rdheader(str(self.get_absolute_path(rec))).sig_name

    def load_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_original: bool = False,
        ignore_beat_types: bool = True,
        extension: str = "q1c",
    ) -> List[ECGWaveForm]:
        """Load the wave delineation annotations of the record.

        The wave delineation annotations are returned
        in the form of list of :class:`ECGWaveForm`.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.
        ignore_beat_types : bool, default True
            If True, the beat types will be ignored (all converted to "N").
        extension : str, default "q1c"
            Extension of the wave delineation file to use.

        Returns
        -------
        wave_list : List[ECGWaveForm]
            The list of wave delineation in the form of :class:`ECGWaveForm`.

        """
        assert extension in [
            "q1c",
            "q2c",
            "pu1",
            "pu2",
        ], "extension should be one of `q1c`, `q2c`, `pu1`, `pu2`"
        fp = str(self.get_absolute_path(rec))
        wfdb_ann = wfdb.rdann(fp, extension=extension)
        header = wfdb.rdheader(fp)
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        subtraction = 0 if keep_original else sf

        wave_list = []
        current_onset = None
        current_wave_name = None
        current_wave_peak = None
        for idx, symbol in zip(wfdb_ann.sample, wfdb_ann.symbol):
            if idx < sf:
                continue
            if idx >= st:
                break
            if symbol == "(":
                current_onset = idx
            elif symbol == ")":
                wave_list.append(
                    ECGWaveForm(
                        onset=(current_onset or np.nan) - subtraction,
                        offset=idx - subtraction,
                        name=current_wave_name,
                        peak=current_wave_peak,
                        duration=(idx - current_onset) / header.fs
                        if current_onset is not None
                        else np.nan,
                    )
                )
                current_onset = None
                current_wave_name = None
                current_wave_peak = None
            else:
                if ignore_beat_types and symbol not in ["p", "t", "u"]:
                    symbol = "N"
                current_wave_name = symbol
                current_wave_peak = idx

        return wave_list

    @add_docstring(load_ann.__doc__)
    def load_wave_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_original: bool = False,
        ignore_beat_types: bool = True,
        extension: str = "q1c",
    ) -> np.ndarray:
        """alias of self.load_ann"""
        return self.load_ann(
            rec,
            sampfrom=sampfrom,
            sampto=sampto,
            keep_original=keep_original,
            ignore_beat_types=ignore_beat_types,
            extension=extension,
        )

    def load_wave_masks(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        mask_format: str = "channel_first",
        class_map: Optional[Dict[str, int]] = None,
        extension: str = "q1c",
    ) -> np.ndarray:
        """Load the wave delineation in the form of masks.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        mask_format : str, default "channel_first"
            Format of the mask,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first").
        class_map : dict, optional
            A custom class map.
            If is None, `self.class_map` will be used.
        extension : str, default "q1c"
            Extension of the wave delineation file to use.

        Returns
        -------
        masks : numpy.ndarray
            The masks corresponding to the wave delineation annotations
            of the record.

        """
        raise NotImplementedError(
            "A large proportion of the wave delineation annotations lack onset indices. "
            "Has to find a rule to give default onset index for the missing ones."
        )

    def load_rhythm_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        rhythm_format: str = "intervals",
        rhythm_types: Optional[Sequence[str]] = None,
        keep_original: bool = False,
        extension: str = "atr",
    ) -> Union[Dict[str, list], np.ndarray]:
        """Load rhythm annotations of a record.

        Rhythm annotations are stored in the `aux_note` attribute
        of corresponding annotation files.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        rhythm_format : {"intervals", "mask"}, optional
            Format of returned annotation, by default "intervals",
            case insensitive.
        rhythm_types : List[str], optional
            The rhythm types to be loaded, defaults to `self.rhythm_types`.
            If is not None, only the rhythm annotations
            with the specified types will be returned.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.
        extension : str, default "atr"
            Extension of the annotation file to use.
            Has to be "atr", since "man" files has no rhythm annotation.

        Returns
        -------
        ann, dict or ndarray,
            the annotations in the format of intervals, or in the format of mask

        """
        raise NotImplementedError(
            "Only a small part of the recordings have rhythm annotations, "
            "hence not implemented yet"
        )

    def load_beat_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        beat_format: str = "beat",
        beat_types: Optional[Sequence[str]] = None,
        keep_original: bool = False,
        extension: str = "atr",
    ) -> Union[Dict[str, np.ndarray], List[BeatAnn]]:
        """Load beat annotations of the record.

        Beat annotations are stored in the `symbol` attribute
        of corresponding annotation files.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        beat_format : {"beat", "dict"}, optional
            Format of returned annotation, by default "beat",
            case insensitive.
        beat_types : List[str], optional
            The beat types to be loaded, defaults to `self.beat_types`.
            If is not None, only the beat annotations
            with the specified types will be returned.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.
        extension : {"atr", "man"}, optional
            Extension of the annotation file, by default "atr",
            case insensitive.

        Returns
        -------
        beat_ann : dict or list
            Locations (indices) of the the given beat types.

        """
        assert beat_format.lower() in [
            "beat",
            "dict",
        ], f"`beat_format` must be one of ['beat', 'dict'], but got `{beat_format}`"
        fp = str(self.get_absolute_path(rec))
        wfdb_ann = wfdb.rdann(fp, extension=extension)
        header = wfdb.rdheader(fp)
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"
        subs = 0 if keep_original else sf

        sample_inds = wfdb_ann.sample
        indices = np.where((sample_inds >= sf) & (sample_inds < st))[0]

        if beat_types is None:
            beat_types = self.beat_types

        beat_ann = [
            BeatAnn(i - subs, s)
            for i, s in zip(sample_inds[indices], np.array(wfdb_ann.symbol)[indices])
            if s in beat_types
        ]

        if beat_format.lower() == "dict":
            beat_ann = {
                s: np.array([b.index for b in beat_ann if b.symbol == s], dtype=int)
                for s in self.beat_types_extended
            }
            beat_ann = {k: v for k, v in beat_ann.items() if len(v) > 0}

        return beat_ann

    def load_rpeak_indices(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_original: bool = False,
        extension: str = "atr",
    ) -> np.ndarray:
        """Load rpeak indices of the record.

        Rpeak indices, or equivalently qrs complex locations,
        which are stored in the `symbol` attribute of corresponding annotation files,
        regardless of their beat types.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.
        extension : {"atr", "man"}, optional
            Extension of the annotation file, by default "atr",
            case insensitive.

        Returns
        -------
        rpeak_inds : numpy.ndarray
            Locations (indices) of the all the rpeaks (qrs complexes).

        """
        assert extension in [
            "atr",
            "man",
        ], f"`extension` must be one of ['atr', 'man'], but got `{extension}`"
        if isinstance(rec, int):
            rec = self[rec]
        rec_fp = self.get_absolute_path(rec)
        if not rec_fp.with_suffix(f".{extension}").exists():
            another_extension = "man" if extension == "atr" else "atr"
            raise FileNotFoundError(
                f"annotation file `{rec_fp.name}` does not exist, "
                f"try setting `extension = \042{another_extension}\042`"
            )
        wfdb_ann = wfdb.rdann(str(rec_fp), extension=extension)
        header = wfdb.rdheader(str(rec_fp))
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        rpeak_inds = wfdb_ann.sample
        indices = np.where(
            (rpeak_inds >= sf)
            & (rpeak_inds < st)
            & (np.isin(wfdb_ann.symbol, self.beat_types))
        )[0]
        rpeak_inds = rpeak_inds[indices]
        if not keep_original:
            rpeak_inds -= sf
        return rpeak_inds

    def plot(
        self,
        rec: Union[str, int],
        data: Optional[np.ndarray] = None,
        ticks_granularity: int = 0,
        leads: Optional[Union[str, int, List[str], List[int]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        same_range: bool = False,
        waves: Optional[ECGWaveForm] = None,
        beat_ann: Optional[Dict[str, np.ndarray]] = None,
        rpeak_inds: Optional[Union[Sequence[int], np.ndarray]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Plot the signals of a record or external signals (units in μV),
        with metadata (fs, labels, tranche, etc.),
        possibly also along with wave delineations.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        data : numpy.ndarray, optional
            The signals to plot.
            If is not None, data of `rec` will not be used.
            This is useful when plotting filtered data
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        leads : str or int or List[str] or List[int], optional
            The leads of the record to plot.
        sampfrom : int, optional
            Start index of the record to plot.
        sampto : int, optional
            End index of the record to plot.
        same_range : bool, default False
            If True, all leads are forced to have the same y range.
        waves : ECGWaveForm, optional
            The waves (p waves, t waves, qrs complexes, etc.).
        beat_ann : dict, optional
            The beat annotations.
        rpeak_inds : numpy.ndarray or List[int], optional
            The rpeak indices.
        kwargs : dict, optional
            Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

        TODO
        ----
        1. Slice too long records, and plot separately for each segment.
        2. Plot waves using :func:`matplotlib.pyplot.axvspan`.

        NOTE
        ----
        `Locator` of ``plt`` has default `MAXTICKS` of 1000.
        If not modifying this number, at most 40 seconds of signal could be plotted once.

        Contributors: Jeethan, and WEN Hao

        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt

            plt.MultipleLocator.MAXTICKS = 3000
        # _leads = self._normalize_leads(leads, standard_ordering=True, lower_cases=False)

        raise NotImplementedError

    @property
    def database_info(self) -> DataBaseInfo:
        return _QTDB_INFO



"""
"""

import re

import numpy as np
import torch
import pytest

from torch_ecg.cfg import CFG
from torch_ecg.augmenters import (
    Augmenter,
    BaselineWanderAugmenter,
    CutMix,
    LabelSmooth,
    Mixup,
    RandomFlip,
    RandomMasking,
    RandomRenormalize,
    StretchCompress,
    StretchCompressOffline,
    AugmenterManager,
)
from torch_ecg.augmenters.baseline_wander import _gen_baseline_wander


SIG_LEN = 2000
BATCH_SIZE = 2
N_LEADS = 12


def test_base_augmenter():
    with pytest.raises(
        TypeError,
        match=f"Can't instantiate abstract class {Augmenter.__name__} with abstract method",
    ):
        aug = Augmenter()


def test_augmenter_manager():
    # all use default config
    config = CFG(
        random=False,
        fs=500,
        baseline_wander={},
        label_smooth={},
        mixup={},
        random_flip={},
        random_masking={},
        random_renormalize={},
        stretch_compress={},
    )
    am = AugmenterManager.from_config(config)

    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    mask1 = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN, 3), dtype=torch.float32)
    mask2 = torch.randint(0, 3, (BATCH_SIZE, SIG_LEN), dtype=torch.long)
    sig, label, mask1, mask2 = am(sig, label, mask1, mask2)

    am.rearrange(
        new_ordering=[
            "stretch_compress",
            "random_masking",
            "baseline_wander",
            "random_renormalize",
            "random_flip",
            "label_smooth",
            "mixup",
        ]
    )

    am.random = True
    sig, label, mask1, mask2 = am(sig, label, mask1, mask2)

    with pytest.warns(
        RuntimeWarning, match="The augmenters are applied in random order"
    ):
        am.random = True
        am.rearrange(
            new_ordering=[
                "mixup",
                "random_masking",
                "random_flip",
                "baseline_wander",
                "random_renormalize",
                "label_smooth",
                "stretch_compress",
            ]
        )
    am.random = False

    with pytest.raises(AssertionError, match="Duplicate augmenter names"):
        am.rearrange(
            new_ordering=[
                "stretch_compress",
                "random_masking",
                "baseline_wander",
                "random_renormalize",
                "random_flip",
                "label_smooth",
                "mixup",
                "random_masking",
            ]
        )

    with pytest.raises(AssertionError, match="Number of augmenters mismatch"):
        am.rearrange(
            new_ordering=[
                "stretch_compress",
                "random_masking",
                "baseline_wander",
                "random_renormalize",
                "random_flip",
                "label_smooth",
            ]
        )

    with pytest.raises(AssertionError, match="Unknown augmenter name: `.+`"):
        am.rearrange(
            new_ordering=[
                "stretch_compress",
                "random_masking",
                "baseline_wander",
                "random_normalize",  # typo
                "random_flip",
                "label_smooth",
                "mixup",
            ]
        )

    assert re.search("augmenters = \\[", repr(am))


def test_baseline_wander_augmenter():
    blw = BaselineWanderAugmenter(300, prob=0.7, inplace=False)
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    sig, _ = blw(sig, label)

    assert str(blw) == repr(blw)

    noise = _gen_baseline_wander(
        siglen=sig.shape[-1],
        fs=blw.fs,
        bw_fs=blw.bw_fs,
        amplitude=blw.ampl_ratio[1],
        amplitude_gaussian=blw.gaussian[1],
    )
    noise = _gen_baseline_wander(
        siglen=sig.shape[-1],
        fs=blw.fs,
        bw_fs=1.5,
        amplitude=0.05,
        amplitude_gaussian=blw.gaussian[1],
    )


def test_cutmix_augmenter():
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    mask = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN, 3), dtype=torch.float32)
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)

    cm = CutMix(fs=500, prob=0.7, beta=0.6)
    sig, label, mask = cm(sig, label, mask)
    assert sig.shape == (BATCH_SIZE, N_LEADS, SIG_LEN)
    assert label.shape == (BATCH_SIZE, 26)
    assert mask.shape == (BATCH_SIZE, SIG_LEN, 3)

    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    mask = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN, 3), dtype=torch.float32)
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)

    cm = CutMix(fs=500, prob=0.4, inplace=False, num_mix=2)
    new_sig, new_label, new_mask = cm(sig, label, mask)
    assert new_sig.shape == (BATCH_SIZE, N_LEADS, SIG_LEN)
    assert new_label.shape == (BATCH_SIZE, 26)
    assert new_mask.shape == (BATCH_SIZE, SIG_LEN, 3)
    new_sig, new_mask = cm(sig, mask)
    assert new_sig.shape == (BATCH_SIZE, N_LEADS, SIG_LEN)
    assert new_mask.shape == (BATCH_SIZE, SIG_LEN, 3)

    assert str(cm) == repr(cm)

    with pytest.raises(
        AssertionError, match="`label` should NOT be categorical labels"
    ):
        label = torch.randint(0, 26, (BATCH_SIZE,), dtype=torch.long)
        cm(sig, label)

    with pytest.raises(
        AssertionError, match="`num_mix` must be a positive integer, but got `.+`"
    ):
        cm = CutMix(fs=500, num_mix=0)
    with pytest.raises(AssertionError, match="Probability must be between 0 and 1"):
        cm = CutMix(fs=500, prob=1.1)
    with pytest.raises(
        AssertionError,
        match="`alpha` and `beta` must be positive, but got `.+` and `.+`",
    ):
        cm = CutMix(fs=500, alpha=0, beta=0)


def test_label_smooth():
    ls = LabelSmooth(inplace=False)
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    _, label = ls(None, label)
    ls = LabelSmooth(smoothing=0.0)
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    _, label = ls(None, label)

    assert str(ls) == repr(ls)


def test_mixup():
    mixup = Mixup()
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    sig, label = mixup(sig, label)
    mixup = Mixup(inplace=False)
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    sig, label = mixup(sig, label)

    assert str(mixup) == repr(mixup)


def test_random_flip():
    rf = RandomFlip()
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    sig, _ = rf(sig, None)
    rf = RandomFlip(inplace=False, per_channel=False)
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    sig, _ = rf(sig, None)
    rf = RandomFlip(prob=0.0, per_channel=False)
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    sig, _ = rf(sig, None)

    assert str(rf) == repr(rf)


def test_random_masking():
    rm = RandomMasking(fs=500, prob=0.7)
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    critical_points = [
        np.arange(250, SIG_LEN - 250, step=400) for _ in range(BATCH_SIZE)
    ]
    sig, _ = rm(sig, None, critical_points=critical_points)
    rm = RandomMasking(fs=500, prob=0.3, inplace=False)
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    sig, _ = rm(sig, None, critical_points=critical_points)
    rm = RandomMasking(fs=500, prob=0.0)
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    sig, _ = rm(sig, None, critical_points=critical_points)

    assert str(rm) == repr(rm)


def test_random_renormalize():
    rrn = RandomRenormalize(per_channel=True, prob=0.7)
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    sig, _ = rrn(sig, None)

    # TODO: fix errors in the following tests
    # rrn = RandomRenormalize(mean=np.zeros((N_LEADS, 1)), std=np.ones((N_LEADS, 1)), per_channel=True)
    # sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    # sig, _ = rrn(sig, None)

    rrn = RandomRenormalize(
        mean=np.zeros((N_LEADS,)), std=np.ones((N_LEADS,)), inplace=False
    )
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    sig, _ = rrn(sig, None)

    rrn = RandomRenormalize(prob=0.0)
    sig = torch.randn(BATCH_SIZE, N_LEADS, SIG_LEN)
    sig, _ = rrn(sig, None)

    assert str(rrn) == repr(rrn)


def test_stretch_compress():
    sc = StretchCompress(inplace=False)
    sig = torch.randn((BATCH_SIZE, N_LEADS, SIG_LEN))
    # labels = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN, 26))
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    mask = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN, 3), dtype=torch.float32)
    sig, label, mask = sc(sig, label, mask)
    assert sig.shape == (BATCH_SIZE, N_LEADS, SIG_LEN)
    assert label.shape == (BATCH_SIZE, 26)
    assert mask.shape == (BATCH_SIZE, SIG_LEN, 3)

    sig = torch.randn((BATCH_SIZE, N_LEADS, SIG_LEN))
    # labels = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN, 26))
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    mask = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN // 8, 3), dtype=torch.float32)
    sig, label, mask = sc._generate(sig, label, mask)

    sc = StretchCompress(prob=0.0)
    sig = torch.randn((BATCH_SIZE, N_LEADS, SIG_LEN))
    # labels = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN, 26))
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    mask = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN // 8, 3), dtype=torch.float32)
    sig, label, mask = sc(sig, label, mask)
    assert sig.shape == (BATCH_SIZE, N_LEADS, SIG_LEN)
    assert label.shape == (BATCH_SIZE, 26)
    assert mask.shape == (BATCH_SIZE, SIG_LEN // 8, 3)

    sig = torch.randn((BATCH_SIZE, N_LEADS, SIG_LEN))
    # labels = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN, 26))
    label = torch.randint(0, 2, (BATCH_SIZE, 26), dtype=torch.float32)
    mask = torch.randint(0, 2, (BATCH_SIZE, SIG_LEN // 8, 3), dtype=torch.float32)
    sig, label, mask = sc._generate(sig, label, mask)

    assert str(sc) == repr(sc)


def test_stretch_compress_offline():
    sco = StretchCompressOffline()
    seglen = 600
    sig = torch.randn((N_LEADS, 60000)).numpy()
    labels = torch.ones((60000, 3)).numpy().astype(int)
    masks = torch.ones((60000, 1)).numpy().astype(int)
    segments = sco(seglen, sig, labels, masks, critical_points=[10000, 30000])

    sig = torch.randn((N_LEADS, 60)).numpy()
    labels = torch.ones((60, 3)).numpy().astype(int)
    masks = torch.ones((60, 1)).numpy().astype(int)
    segments = sco(seglen, sig, labels, masks)

    sig = torch.randn((N_LEADS, 800)).numpy()
    labels = torch.ones((800, 3)).numpy().astype(int)
    masks = torch.ones((800, 1)).numpy().astype(int)
    segments = sco(seglen, sig, labels, masks)

    assert str(sco) == repr(sco)



"""
torch_ecg.augmenters
====================

Augmenters are used to augment the training data.

.. contents:: torch_ecg.augmenters
    :depth: 2
    :local:
    :backlinks: top

.. currentmodule:: torch_ecg.augmenters

.. autosummary::
    :toctree: generated/
    :recursive:

    AugmenterManager
    Augmenter
    BaselineWanderAugmenter
    CutMix
    LabelSmooth
    Mixup
    RandomFlip
    RandomMasking
    RandomRenormalize
    StretchCompress
    StretchCompressOffline

"""

from .augmenter_manager import AugmenterManager
from .base import Augmenter
from .baseline_wander import BaselineWanderAugmenter
from .cutmix import CutMix
from .label_smooth import LabelSmooth
from .mixup import Mixup
from .random_flip import RandomFlip
from .random_masking import RandomMasking
from .random_renormalize import RandomRenormalize
from .stretch_compress import StretchCompress, StretchCompressOffline


__all__ = [
    "Augmenter",
    "BaselineWanderAugmenter",
    "CutMix",
    "LabelSmooth",
    "Mixup",
    "RandomFlip",
    "RandomMasking",
    "RandomRenormalize",
    "StretchCompress",
    "StretchCompressOffline",
    "AugmenterManager",
]



"""
torch_ecg.preprocessors
=======================

This module contains the preprocessors for signals of torch Tensor format.

.. contents:: torch_ecg.preprocessors
    :depth: 2
    :local:
    :backlinks: top

.. currentmodule:: torch_ecg.preprocessors

.. autosummary::
    :toctree: generated/
    :recursive:

    PreprocManager
    BandPass
    BaselineRemove
    Normalize
    MinMaxNormalize
    NaiveNormalize
    ZScoreNormalize
    Resample

"""

from .bandpass import BandPass
from .baseline_remove import BaselineRemove
from .normalize import MinMaxNormalize, NaiveNormalize, Normalize, ZScoreNormalize
from .preproc_manager import PreprocManager
from .resample import Resample


__all__ = [
    "PreprocManager",
    "BandPass",
    "BaselineRemove",
    "Normalize",
    "MinMaxNormalize",
    "NaiveNormalize",
    "ZScoreNormalize",
    "Resample",
]



"""
"""

import warnings
from numbers import Real
from typing import Any

import torch

from .._preprocessors.base import preprocess_multi_lead_signal


__all__ = [
    "BaselineRemove",
]


class BaselineRemove(torch.nn.Module):
    """Baseline removal using median filtering.

    Parameters
    ----------
    fs : numbers.Real
        Sampling frequency of the ECG signal to be filtered.
    window1 : float, default 0.2
        The smaller window size of the median filter,
        with units in seconds.
    window2 : float, default 0.6
        The larger window size of the median filter,
        with units in seconds.
    inplace : bool, default True
        Whether to perform the filtering in-place.
    kwargs : dict, optional
        Other keyword arguments for :class:`torch.nn.Module`.

    """

    __name__ = "BaselineRemove"

    def __init__(
        self,
        fs: Real,
        window1: float = 0.2,
        window2: float = 0.6,
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.fs = fs
        self.window1 = window1
        self.window2 = window2
        if self.window2 < self.window1:
            self.window1, self.window2 = self.window2, self.window1
            warnings.warn(
                "values of `window1` and `window2` are switched", RuntimeWarning
            )
        self.inplace = inplace

    def forward(self, sig: torch.Tensor) -> torch.Tensor:
        """Apply the preprocessor to the signal tensor.

        Parameters
        ----------
        sig : torch.Tensor
            The ECG signal tensor,
            of shape ``(batch, lead, siglen)``.

        Returns
        -------
        torch.Tensor
            The median filtered (hence baseline removed) ECG signals,
            of shape ``(batch, lead, siglen)``.

        """
        if not self.inplace:
            sig = sig.clone()
        sig = torch.as_tensor(
            preprocess_multi_lead_signal(
                raw_sig=sig.cpu().numpy(),
                fs=self.fs,
                bl_win=[self.window1, self.window2],
            ).copy(),
            dtype=sig.dtype,
            device=sig.device,
        )
        return sig



"""
utilities for nn models

"""

import re
import warnings
from copy import deepcopy
from itertools import repeat, chain
from math import floor
from numbers import Real
from pathlib import Path
from typing import List, Optional, Sequence, Tuple, Union, Dict

import numpy as np
import torch
from torch import Tensor, nn

from ..cfg import CFG, DEFAULTS
from .misc import add_docstring
from .utils_data import cls_to_bin


__all__ = [
    "extend_predictions",
    "compute_output_shape",
    "compute_conv_output_shape",
    "compute_deconv_output_shape",
    "compute_maxpool_output_shape",
    "compute_avgpool_output_shape",
    "compute_sequential_output_shape",
    "compute_sequential_output_shape_docstring",
    "compute_module_size",
    "default_collate_fn",
    "compute_receptive_field",
    "adjust_cnn_filter_lengths",
    "SizeMixin",
    "CkptMixin",
]


def extend_predictions(
    preds: Sequence, classes: List[str], extended_classes: List[str]
) -> np.ndarray:
    """Extend the prediction arrays to prediction arrays in larger range of classes

    Parameters
    ----------
    preds : array_like
        Array of predictions (scalar or binary) of shape
        ``(n_records, n_classes)``,
        or categorical predictions of shape ``(n_classes,)``,
        where ``n_classes = len(classes)``.
    classes : List[str]
        Classes of the predictions of `preds`.
    extended_classes : List[str]
        A superset of `classes`.
        The predictions will be extended to this range of classes.

    Returns
    -------
    extended_preds : numpy.ndarray
        The extended array of predictions, with indices in `extended_classes`,
        of shape ``(n_records, n_classes)``, or ``(n_classes,)``.

    Examples
    --------
    .. code-block:: python

        n_records, n_classes = 10, 3
        classes = ["NSR", "AF", "PVC"]
        extended_classes = ["AF", "RBBB", "PVC", "NSR"]
        scalar_pred = torch.rand(n_records, n_classes)
        extended_pred = extend_predictions(scalar_pred, classes, extended_classes)
        bin_pred = torch.randint(0, 2, (n_records, n_classes))
        extended_pred = extend_predictions(bin_pred, classes, extended_classes)
        cate_pred = torch.randint(0, n_classes, (n_records,))
        extended_pred = extend_predictions(cate_pred, classes, extended_classes)

    """
    assert len(set(classes) - set(extended_classes)) == 0, (
        "`extended_classes` is not a superset of `classes`, "
        f"with {set(classes)-set(extended_classes)} in `classes` but not in `extended_classes`"
    )

    if isinstance(preds, Tensor):
        _preds = preds.numpy()
    else:
        _preds = np.array(preds)

    if np.ndim(_preds) == 1:  # categorical predictions
        extended_preds = cls_to_bin(_preds, len(classes))
        extended_preds = extend_predictions(extended_preds, classes, extended_classes)
        extended_preds = np.where(extended_preds == 1)[1]
        return extended_preds

    assert _preds.shape[1] == len(
        classes
    ), f"`pred` indicates {_preds.shape[1]} classes, while `classes` has {len(classes)}"

    extended_preds = np.zeros((_preds.shape[0], len(extended_classes)))

    for idx, c in enumerate(classes):
        new_idx = extended_classes.index(c)
        extended_preds[..., new_idx] = _preds[..., idx]

    if np.array(preds).ndim == 1:
        extended_preds = extended_preds[0]

    return extended_preds


# utils for computing output shape
def compute_output_shape(
    layer_type: str,
    input_shape: Sequence[Union[int, None]],
    num_filters: Optional[int] = None,
    kernel_size: Union[Sequence[int], int] = 1,
    stride: Union[Sequence[int], int] = 1,
    padding: Union[Sequence[int], int] = 0,
    output_padding: Union[Sequence[int], int] = 0,
    dilation: Union[Sequence[int], int] = 1,
    channel_last: bool = False,
    asymmetric_padding: Union[Sequence[int], Sequence[Sequence[int]]] = None,
) -> Tuple[Union[int, None]]:
    """Compute the output shape of a (transpose) convolution/maxpool/avgpool layer.

    This function is based on the discussion [#disc]_.

    Parameters
    ----------
    layer_type : str
        Type (conv, maxpool, avgpool, etc.) of the layer.
    input_shape : Sequence[Union[int, None]]
        Shape of an input :class:`~torch.Tensor`.
        The first dimension is the batch dimension,
        which is allowed to be `None`.
    num_filters : int, optional
        Number of filters, also the channel dimension.
    kernel_size : int or Sequence[int], default 1
        Kernel size (filter size) of the layer,
        should be compatible with `input_shape`.
    stride : int or Sequence[int], default 1
        Stride (down-sampling length) of the layer,
        should be compatible with `input_shape`.
    padding : int or Sequence[int], default 0
        Padding length(s) of the layer,
        should be compatible with `input_shape`.
    output_padding : int or Sequence[int], default 0
        Additional size added to one side of the output shape,
        used only for transpose convolution.
    dilation : int or Sequence[int], default 1
        Dilation of the layer, should be compatible with `input_shape`.
    channel_last : bool, default False
        Whether the channel dimension is the last dimension,
        or the second dimension (the first is the batch dimension by convention).
    asymmetric_padding : Sequence[int] or Sequence[Sequence[int]], optional
        (2-)sequence of int or sequence of (2-)sequence of int
        asymmetric paddings for all dimensions or for each dimension.

    Returns
    -------
    output_shape : tuple
        Shape of the output :class:`~torch.Tensor`.

    References
    ----------
    .. [#disc] https://discuss.pytorch.org/t/utility-function-for-calculating-the-shape-of-a-conv-output/11173/5

    """
    # check validity of arguments
    __TYPES__ = [
        "conv",
        "convolution",
        "deconv",
        "deconvolution",
        "transposeconv",
        "transposeconvolution",
        "maxpool",
        "maxpooling",
        "avgpool",
        "avgpooling",
        "averagepool",
        "averagepooling",
    ]
    lt = "".join(layer_type.lower().split("_"))
    assert (
        lt in __TYPES__
    ), f"Unknown layer type `{layer_type}`, should be one of: {__TYPES__}"

    def assert_positive_integer(num):
        return isinstance(num, int) and num > 0

    def assert_non_negative_integer(num):
        return isinstance(num, int) and num >= 0

    dim = len(input_shape) - 2
    assert dim > 0, (
        "`input_shape` should be a sequence of length at least 3, "
        "to be a valid (with batch and channel) shape of a non-degenerate Tensor"
    )
    assert all(
        s is None or assert_positive_integer(s) for s in input_shape
    ), "`input_shape` should be a sequence containing only `None` and positive integers"

    if num_filters is not None:
        assert assert_positive_integer(
            num_filters
        ), "`num_filters` should be `None` or positive integer"
    assert all(
        assert_positive_integer(num)
        for num in np.asarray(kernel_size).flatten().tolist()
    ), "`kernel_size` should contain only positive integers"
    assert all(
        assert_positive_integer(num) for num in np.asarray(stride).flatten().tolist()
    ), "`stride` should contain only positive integers"
    assert all(
        assert_non_negative_integer(num)
        for num in np.asarray(padding).flatten().tolist()
    ), "`padding` should contain only non-negative integers"
    assert all(
        assert_non_negative_integer(num)
        for num in np.asarray(output_padding).flatten().tolist()
    ), "`output_padding` should contain only non-negative integers"
    assert all(
        assert_positive_integer(num) for num in np.asarray(dilation).flatten().tolist()
    ), "`dilation` should contain only positive integers"

    if lt in [
        "conv",
        "convolution",
    ]:
        # as function of dilation, kernel_size
        def minus_term(d, k):
            return d * (k - 1) + 1

        out_channels = num_filters
    elif lt in [
        "maxpool",
        "maxpooling",
    ]:

        def minus_term(d, k):
            return d * (k - 1) + 1

        out_channels = input_shape[-1] if channel_last else input_shape[1]
    elif lt in [
        "avgpool",
        "avgpooling",
        "averagepool",
        "averagepooling",
    ]:

        def minus_term(d, k):
            return k

        out_channels = input_shape[-1] if channel_last else input_shape[1]
    elif lt in [
        "deconv",
        "deconvolution",
        "transposeconv",
        "transposeconvolution",
    ]:
        out_channels = num_filters

    def check_output_validity(shape):
        assert all(
            p is None or p > 0 for p in shape
        ), f"output shape `{shape}` is illegal, please check input arguments"
        return shape

    # none_dim_msg = "only batch and channel dimension can be `None`"
    # if channel_last:
    #     assert all([n is not None for n in input_shape[1:-1]]), none_dim_msg
    # else:
    #     assert all([n is not None for n in input_shape[2:]]), none_dim_msg
    none_dim_msg = "spatial dimensions should be all `None`, or all not `None`"
    if channel_last:
        if all([n is None for n in input_shape[1:-1]]):
            if out_channels is None:
                raise ValueError(
                    "out channel dimension and spatial dimensions are all `None`"
                )
            output_shape = tuple(list(input_shape[:-1]) + [out_channels])
            return check_output_validity(output_shape)
        elif any([n is None for n in input_shape[1:-1]]):
            raise ValueError(none_dim_msg)
    else:
        if all([n is None for n in input_shape[2:]]):
            if out_channels is None:
                raise ValueError(
                    "out channel dimension and spatial dimensions are all `None`"
                )
            output_shape = tuple([input_shape[0], out_channels] + list(input_shape[2:]))
            return check_output_validity(output_shape)
        elif any([n is None for n in input_shape[2:]]):
            raise ValueError(none_dim_msg)

    if isinstance(kernel_size, int):
        _kernel_size = list(repeat(kernel_size, dim))
    elif len(kernel_size) == dim:
        _kernel_size = kernel_size
    else:
        raise ValueError(
            f"input has {dim} dimensions, while kernel has {len(kernel_size)} dimensions, "
            "both not including the channel dimension"
        )

    if isinstance(stride, int):
        _stride = list(repeat(stride, dim))
    elif len(stride) == dim:
        _stride = stride
    else:
        raise ValueError(
            f"input has {dim} dimensions, while `kernel` has {len(stride)} dimensions, "
            "both not including the channel dimension"
        )

    # NOTE: asymmetric padding along one spatial dimension
    # seems not supported yet by PyTorch's builtin Module classes
    if isinstance(padding, int):
        _padding = list(repeat(list(repeat(padding, 2)), dim))
    # elif len(padding) == 2 and isinstance(padding[0], int):
    #     _padding = list(repeat(padding, dim))
    # elif (
    #     len(padding) == dim
    #     and all([isinstance(p, Sequence) for p in padding])
    #     and all([len(p) == 2 for p in padding])
    # ):
    elif len(padding) == dim:
        _padding = [list(repeat(p, 2)) for p in padding]
    else:
        raise ValueError(
            f"input has {dim} dimensions, while `padding` has {len(padding)} dimensions, "
            "both not including the channel dimension"
        )
        # raise ValueError("Invalid `padding`")

    if asymmetric_padding is not None:
        assert hasattr(asymmetric_padding, "__len__"), "Invalid `asymmetric_padding`"
        if isinstance(asymmetric_padding[0], int):
            assert len(asymmetric_padding) == 2 and isinstance(
                asymmetric_padding[1], int
            ), "Invalid `asymmetric_padding`"
            _asymmetric_padding = list(repeat(asymmetric_padding, dim))
        else:
            assert len(asymmetric_padding) == dim and all(
                len(ap) == 2 and all(isinstance(p, int) for p in ap)
                for ap in asymmetric_padding
            ), "Invalid `asymmetric_padding`"
            _asymmetric_padding = asymmetric_padding
        for idx in range(dim):
            _padding[idx][0] += _asymmetric_padding[idx][0]
            _padding[idx][1] += _asymmetric_padding[idx][1]

    if isinstance(output_padding, int):
        _output_padding = list(repeat(output_padding, dim))
    elif len(output_padding) == dim:
        _output_padding = output_padding
    else:
        raise ValueError(
            f"input has {dim} dimensions, while `output_padding` has {len(output_padding)} dimensions, "
            "both not including the channel dimension"
        )

    if isinstance(dilation, int):
        _dilation = list(repeat(dilation, dim))
    elif len(dilation) == dim:
        _dilation = dilation
    else:
        raise ValueError(
            f"input has {dim} dimensions, while `dilation` has {len(dilation)} dimensions, "
            "both not including the channel dimension"
        )

    if channel_last:
        _input_shape = list(input_shape[1:-1])
    else:
        _input_shape = list(input_shape[2:])

    if lt in [
        "deconv",
        "deconvolution",
        "transposeconv",
        "transposeconvolution",
    ]:
        output_shape = [
            (i - 1) * s - sum(p) + d * (k - 1) + o + 1
            for i, p, o, d, k, s in zip(
                _input_shape,
                _padding,
                _output_padding,
                _dilation,
                _kernel_size,
                _stride,
            )
        ]
    else:
        output_shape = [
            floor(((i + sum(p) - minus_term(d, k)) / s) + 1)
            for i, p, d, k, s in zip(
                _input_shape, _padding, _dilation, _kernel_size, _stride
            )
        ]
    if channel_last:
        output_shape = tuple([input_shape[0]] + output_shape + [out_channels])
    else:
        output_shape = tuple([input_shape[0], out_channels] + output_shape)

    return check_output_validity(output_shape)


def compute_conv_output_shape(
    input_shape: Sequence[Union[int, None]],
    num_filters: Optional[int] = None,
    kernel_size: Union[Sequence[int], int] = 1,
    stride: Union[Sequence[int], int] = 1,
    padding: Union[Sequence[int], int] = 0,
    dilation: Union[Sequence[int], int] = 1,
    channel_last: bool = False,
    asymmetric_padding: Union[Sequence[int], Sequence[Sequence[int]]] = None,
) -> Tuple[Union[int, None]]:
    """Compute the output shape of a convolution layer.

    Parameters
    ----------
    input_shape : Sequence[Union[int, None]]
        Shape of an input :class:`~torch.Tensor`.
        The first dimension is the batch dimension,
        which is allowed to be `None`.
    num_filters : int, optional
        Number of filters, also the channel dimension.
    kernel_size : int or Sequence[int], default 1
        Kernel size (filter size) of the layer,
        should be compatible with `input_shape`.
    stride : int or Sequence[int], default 1
        Stride (down-sampling length) of the layer,
        should be compatible with `input_shape`.
    padding : int or Sequence[int], default 0
        Padding length(s) of the layer,
        should be compatible with `input_shape`.
    dilation : int or Sequence[int], default 1
        Dilation of the layer, should be compatible with `input_shape`.
    channel_last : bool, default False
        Whether the channel dimension is the last dimension,
        or the second dimension (the first is the batch dimension by convention).
    asymmetric_padding : Sequence[int] or Sequence[Sequence[int]], optional
        (2-)sequence of int or sequence of (2-)sequence of int
        asymmetric paddings for all dimensions or for each dimension.

    Returns
    -------
    output_shape : tuple
        Shape of the output :class:`~torch.Tensor`.

    """
    output_shape = compute_output_shape(
        "conv",
        input_shape,
        num_filters,
        kernel_size,
        stride,
        padding,
        0,
        dilation,
        channel_last,
        asymmetric_padding,
    )
    return output_shape


def compute_maxpool_output_shape(
    input_shape: Sequence[Union[int, None]],
    kernel_size: Union[Sequence[int], int] = 1,
    stride: Union[Sequence[int], int] = 1,
    padding: Union[Sequence[int], int] = 0,
    dilation: Union[Sequence[int], int] = 1,
    channel_last: bool = False,
) -> Tuple[Union[int, None]]:
    """Compute the output shape of a maxpool layer.

    Parameters
    ----------
    input_shape : Sequence[Union[int, None]]
        Shape of an input :class:`~torch.Tensor`.
        The first dimension is the batch dimension,
        which is allowed to be `None`.
    kernel_size : int or Sequence[int], default 1
        Kernel size (filter size) of the layer,
        should be compatible with `input_shape`.
    stride : int or Sequence[int], default 1
        Stride (down-sampling length) of the layer,
        should be compatible with `input_shape`.
    padding : int or Sequence[int], default 0
        Padding length(s) of the layer,
        should be compatible with `input_shape`.
    dilation : int or Sequence[int], default 1
        Dilation of the layer, should be compatible with `input_shape`.
    channel_last : bool, default False
        Whether the channel dimension is the last dimension,
        or the second dimension (the first is the batch dimension by convention).

    Returns
    -------
    output_shape : tuple
        Shape of the output :class:`~torch.Tensor`.

    """
    output_shape = compute_output_shape(
        "maxpool",
        input_shape,
        1,
        kernel_size,
        stride,
        padding,
        0,
        dilation,
        channel_last,
    )
    return output_shape


def compute_avgpool_output_shape(
    input_shape: Sequence[Union[int, None]],
    kernel_size: Union[Sequence[int], int] = 1,
    stride: Union[Sequence[int], int] = 1,
    padding: Union[Sequence[int], int] = 0,
    channel_last: bool = False,
) -> Tuple[Union[int, None]]:
    """Compute the output shape of a avgpool layer.

    Parameters
    ----------
    input_shape : Sequence[Union[int, None]]
        Shape of an input :class:`~torch.Tensor`.
        The first dimension is the batch dimension,
        which is allowed to be `None`.
    kernel_size : int or Sequence[int], default 1
        Kernel size (filter size) of the layer,
        should be compatible with `input_shape`.
    stride : int or Sequence[int], default 1
        Stride (down-sampling length) of the layer,
        should be compatible with `input_shape`.
    padding : int or Sequence[int], default 0
        Padding length(s) of the layer,
        should be compatible with `input_shape`.
    channel_last : bool, default False
        Whether the channel dimension is the last dimension,
        or the second dimension (the first is the batch dimension by convention).

    Returns
    -------
    output_shape : tuple
        Shape of the output :class:`~torch.Tensor`.

    """
    output_shape = compute_output_shape(
        "avgpool",
        input_shape,
        1,
        kernel_size,
        stride,
        padding,
        0,
        1,
        channel_last,
    )
    return output_shape


def compute_deconv_output_shape(
    input_shape: Sequence[Union[int, None]],
    num_filters: Optional[int] = None,
    kernel_size: Union[Sequence[int], int] = 1,
    stride: Union[Sequence[int], int] = 1,
    padding: Union[Sequence[int], int] = 0,
    output_padding: Union[Sequence[int], int] = 0,
    dilation: Union[Sequence[int], int] = 1,
    channel_last: bool = False,
    asymmetric_padding: Union[Sequence[int], Sequence[Sequence[int]]] = None,
) -> Tuple[Union[int, None]]:
    """Compute the output shape of a transpose convolution layer

    Parameters
    ----------
    input_shape : Sequence[Union[int, None]]
        Shape of an input :class:`~torch.Tensor`.
        The first dimension is the batch dimension,
        which is allowed to be `None`.
    num_filters : int, optional
        Number of filters, also the channel dimension.
    kernel_size : int or Sequence[int], default 1
        Kernel size (filter size) of the layer,
        should be compatible with `input_shape`.
    stride : int or Sequence[int], default 1
        Stride (down-sampling length) of the layer,
        should be compatible with `input_shape`.
    padding : int or Sequence[int], default 0
        Padding length(s) of the layer,
        should be compatible with `input_shape`.
    output_padding : int or Sequence[int], default 0
        Additional size added to one side of the output shape.
    dilation : int or Sequence[int], default 1
        Dilation of the layer, should be compatible with `input_shape`.
    channel_last : bool, default False
        Whether the channel dimension is the last dimension,
        or the second dimension (the first is the batch dimension by convention).
    asymmetric_padding : Sequence[int] or Sequence[Sequence[int]], optional
        (2-)sequence of int or sequence of (2-)sequence of int
        asymmetric paddings for all dimensions or for each dimension.

    Returns
    -------
    output_shape : tuple
        Shape of the output :class:`~torch.Tensor`.

    """
    output_shape = compute_output_shape(
        "deconv",
        input_shape,
        num_filters,
        kernel_size,
        stride,
        padding,
        output_padding,
        dilation,
        channel_last,
        asymmetric_padding,
    )
    return output_shape


compute_sequential_output_shape_docstring = """

    Parameters
    ----------
    seq_len : int, optional
        Length of the input tensors.
    batch_size : int, optional
        Batch size of the input tensors.

    Returns
    -------
    output_shape : sequence
        The output shape of the module.

    """


@add_docstring(compute_sequential_output_shape_docstring, mode="append")
def compute_sequential_output_shape(
    model: nn.Sequential,
    seq_len: Optional[int] = None,
    batch_size: Optional[int] = None,
) -> Sequence[Union[int, None]]:
    """Compute the output shape of a sequential model."""
    assert issubclass(
        type(model), nn.Sequential
    ), f"model should be nn.Sequential, but got {type(model)}"
    _seq_len = seq_len
    for module in model:
        output_shape = module.compute_output_shape(_seq_len, batch_size)
        _, _, _seq_len = output_shape
    return output_shape


def compute_module_size(
    module: nn.Module,
    requires_grad: bool = True,
    include_buffers: bool = False,
    human: bool = False,
) -> Union[int, str]:
    """compute the size (number of parameters)
    of a :class:`~torch.nn.Module`.

    Parameters
    ----------
    module : torch.nn.Module
        The :class:`~torch.nn.Module` to compute the size.
    requires_grad : bool, default True
        Whether to only count the parameters that require gradients.
    include_buffers : bool, default False
        Whether to include the buffers.
        If `requires_grad` is True, then `include_buffers` is ignored.
    human : bool, default False
        Size is returned in a way that is easy to read by a human,
        by appending a suffix corresponding to the unit (B, K, M, G, T, P).

    Returns
    -------
    n_params : int or str
        Size (number of parameters) of this :class:`~torch.nn.Module`,
        or a string representing the memory size.

    Examples
    --------
    >>> import torch
    >>> class Model(torch.nn.Sequential):
    ...     def __init__(self):
    ...         super().__init__()
    ...         self.add_module("linear", torch.nn.Linear(10, 20, dtype=torch.float16))
    ...         self.register_buffer("hehe", torch.ones(20, 2, dtype=torch.float64))
    >>> model = Model()
    >>> model.linear.weight.requires_grad_(False)
    >>> compute_module_size(model)
    20
    >>> compute_module_size(model, requires_grad=False)
    220
    >>> compute_module_size(model, requires_grad=False, include_buffers=True)
    260
    >>> compute_module_size(model, requires_grad=False, include_buffers=True, human=True)
    '0.7K'
    >>> compute_module_size(model, requires_grad=False, include_buffers=False, human=True)
    '0.4K'
    >>> compute_module_size(model, human=True)
    '40.0B'

    """
    if requires_grad:
        tensor_containers = filter(lambda p: p.requires_grad, module.parameters())
        if include_buffers:
            warnings.warn(
                "`include_buffers` is ignored when `requires_grad` is True",
                RuntimeWarning,
            )
    elif include_buffers:
        tensor_containers = chain(module.parameters(), module.buffers())
    else:
        tensor_containers = module.parameters()
    if human:
        size_dict = {
            "torch.float16": 2,
            "torch.float32": 4,
            "torch.float64": 8,
            "torch.int8": 1,
            "torch.int16": 2,
            "torch.int32": 4,
            "torch.int64": 8,
            "torch.uint8": 1,
        }
        n_params = sum(
            [
                np.prod(item.size()) * size_dict[str(item.dtype)]
                for item in tensor_containers
            ]
        )
        div_count = 0
        while n_params >= 1024 * 0.1:
            n_params /= 1024
            div_count += 1
        cvt_dict = {c: u for c, u in enumerate(list("BKMGTP"))}
        n_params = f"""{n_params:.1f}{cvt_dict[div_count]}"""
    else:
        n_params = int(sum([np.prod(item.size()) for item in tensor_containers]))
    return n_params


def compute_receptive_field(
    kernel_sizes: Union[Sequence[int], int] = 1,
    strides: Union[Sequence[int], int] = 1,
    dilations: Union[Sequence[int], int] = 1,
    input_len: Optional[int] = None,
) -> Union[int, float]:
    """Compute the receptive field of several types of
    :class:`~torch.nn.Module`.

    Computes the (generic) receptive field of feature map of certain channel,
    from certain flow (if not merged, different flows, e.g.
    shortcut, must be computed separately),
    for convolutions, (non-global) poolings.
    "generic" refers to a general position, rather than specific positions,
    like on the edges, whose receptive field is definitely different.

    .. math::

        Let the layers has kernel size, stride, dilation $(k_n, s_n, d_n)$ respectively.
        Let each feature map has receptive field length $r_n$,
        and difference of receptive fields of adjacent positions be $f_n$.
        By convention, $(r_0, f_0) = (1, 1)$. Then one has
        \\begin{eqnarray}
        r_{n+1} & = & r_n + d_n(k_n-1)f_n, \\\\
        f_{n+1} & = & s_n f_n.
        \\end{eqnarray}
        Hence
        \\begin{eqnarray}
        f_{n} & = & \\prod\\limits_{i=0}^{n-1} s_i, \\\\
        r_{n} & = & 1 + \\sum\\limits_{i=0}^{n-1} d_i(k_i-1) \\prod\\limits_{i=0}^{j-1} s_j,
        \\end{eqnarray}
        with empty products equaling 1 by convention.

    Parameters
    ----------
    kernel_sizes : int or Sequence[int], default 1
        The sequence of kernel size for all the layers in the flow.
    strides : int or Sequence[int], default 1
        The sequence of strides for all the layers in the flow
    dilations : int or Sequence[int], default 1
        The sequence of strides for all the layers in the flow
    input_len : int, optional
        Length of the first feature map in the flow.

    Returns
    -------
    receptive_field : int or float
        (Length) of the receptive field.

    Examples
    --------
    >>> compute_receptive_field([11,2,7,7,2,5,5,5,2],[1,2,1,1,2,1,1,1,2])
    90
    >>> compute_receptive_field([11,2,7,7,2,5,5,5,2],[1,2,1,1,2,1,1,1,2],[2,1,2,4,1,8,8,8,1])
    484
    >>> compute_receptive_field([11,2,7,7,2,5,5,5,2],[1,2,1,1,2,1,1,1,2],[4,1,4,8,1,16,32,64,1])
    1984

    The above example exhibits the receptive fields of the output feature maps
    of the 3 branches of the multi-scopic net, using its original hyper-parameters,
    (note the 3 max pooling layers).

    """
    _kernel_sizes = (
        [kernel_sizes] if isinstance(kernel_sizes, int) else list(kernel_sizes)
    )
    num_layers = len(_kernel_sizes)
    _strides = (
        list(repeat(strides, num_layers)) if isinstance(strides, int) else list(strides)
    )
    _dilations = (
        list(repeat(dilations, num_layers))
        if isinstance(dilations, int)
        else list(dilations)
    )
    assert num_layers == len(_strides) == len(_dilations)
    receptive_field = 1
    for idx, (k, d) in enumerate(zip(_kernel_sizes, _dilations)):
        s = np.prod(_strides[:idx]) if idx > 0 else 1
        receptive_field += d * (k - 1) * s
    if input_len is not None:
        receptive_field = min(receptive_field, input_len)
    return receptive_field


def default_collate_fn(
    batch: Sequence[Union[Tuple[np.ndarray, ...], Dict[str, np.ndarray]]]
) -> Union[Tuple[Tensor, ...], Dict[str, Tensor]]:
    """Default collate functions for model training.

    The data generator (:class:`~torch.utils.data.Dataset`) should
    generate (`__getitem__`) n-tuples ``signals, labels, ...``,
    or dictionaries of tensors.

    Parameters
    ----------
    batch : sequence
        Sequence of n-tuples,
        in which the first element is the signal, the second is the label, ...;
        or sequence of dictionaries of tensors.

    Returns
    -------
    tuple or dict
        Tuple or dict of :class:`~torch.Tensor`, which are
        the concatenated values to feed into neural networks.

    """
    if isinstance(batch[0], dict):
        keys = batch[0].keys()
        collated = _default_collate_fn([tuple(b[k] for k in keys) for b in batch])
        return {k: collated[i] for i, k in enumerate(keys)}
    else:
        return _default_collate_fn(batch)


def _default_collate_fn(batch: Sequence[Tuple[np.ndarray, ...]]) -> Tuple[Tensor, ...]:
    """Collate functions for tuples of tensors.

    The data generator (:class:`~torch.utils.data.Dataset`) should
    generate (`__getitem__`) n-tuples ``signals, labels, ...``,
    or dictionaries of tensors.

    Parameters
    ----------
    batch : sequence
        Sequence of n-tuples,
        in which the first element is the signal, the second is the label, ...

    Returns
    -------
    Tuple[torch.Tensor]
        The concatenated values to feed into neural networks.

    """
    try:
        n_fields = len(batch[0])
    except Exception:
        raise ValueError("Invalid batch")
    if n_fields == 0:
        raise ValueError("No data")
    ret = []
    for i in range(n_fields):
        values = [[item[i]] for item in batch]
        values = np.concatenate(values, axis=0)
        values = torch.from_numpy(values)
        ret.append(values)
    return tuple(ret)


def _adjust_cnn_filter_lengths(
    config: dict,
    fs: int,
    ensure_odd: bool = True,
    pattern: str = "filter_length|filter_size",
) -> dict:
    """
    Adjust the filter lengths (kernel sizes) in the config for
    convolutional neural networks, according to the new sampling frequency.

    Parameters
    ----------
    config : dict
        The config dictionary. This `dict` is **NOT** modified.
    fs : int
        The new sampling frequency.
    ensure_odd : bool, default True
        If True, the new filter lengths are ensured to be odd.
    pattern : str, default "filter_length|filter_size"
        The pattern to search for in the config items related to filter lengths.

    Returns
    -------
    config : dict
        The adjusted config dictionary.

    """
    assert "fs" in config
    config = deepcopy(config)
    for k, v in config.items():
        if isinstance(v, dict):
            tmp_config = deepcopy(v)
            original_fs = tmp_config.get("fs", None)
            tmp_config.update({"fs": config["fs"]})
            config[k] = _adjust_cnn_filter_lengths(tmp_config, fs, ensure_odd, pattern)
            if original_fs is None:
                config[k].pop("fs", None)
            else:
                config[k]["fs"] = fs
        elif re.findall(pattern, k):
            if isinstance(v, (Sequence, np.ndarray)):  # DO NOT use `Iterable`
                config[k] = [
                    _adjust_cnn_filter_lengths(
                        {"filter_length": fl, "fs": config["fs"]}, fs, ensure_odd
                    )["filter_length"]
                    for fl in v
                ]
            elif isinstance(v, Real):
                # DO NOT use `int`, which might not work for numpy array elements
                if v > 1:
                    config[k] = int(round(v * fs / config["fs"]))
                    if ensure_odd:
                        config[k] = config[k] - config[k] % 2 + 1
        elif isinstance(v, Sequence) and not isinstance(v, (str, bytes)):
            tmp_configs = [
                _adjust_cnn_filter_lengths(
                    {k: item, "fs": config["fs"]}, fs, ensure_odd, pattern
                )
                for item in v
            ]
            config[k] = [item[k] for item in tmp_configs]
    return config


def adjust_cnn_filter_lengths(
    config: dict,
    fs: int,
    ensure_odd: bool = True,
    pattern: str = "filter_length|filter_size",
) -> dict:
    """
    Adjust the filter lengths in the config for
    convolutional neural networks, according to the new sampling frequency.

    Parameters
    ----------
    config : dict
        The config dictionary. This `dict` is **NOT** modified.
    fs : int
        The new sampling frequency.
    ensure_odd : bool, default True
        If True, the new filter lengths are ensured to be odd.
    pattern : str, default "filter_length|filter_size"
        The pattern to search for in the config items related to filter lengths.

    Returns
    -------
    config : dict
        The adjusted config dictionary.

    """
    config = _adjust_cnn_filter_lengths(config, fs, ensure_odd, pattern)
    config["fs"] = fs
    return config


class SizeMixin(object):
    """Mixin class for size related methods"""

    @property
    def module_size(self) -> int:
        return compute_module_size(self)

    @property
    def module_size_(self) -> str:
        return compute_module_size(self, human=True)

    @property
    def sizeof(self) -> int:
        return compute_module_size(
            self, requires_grad=False, include_buffers=True, human=False
        )

    @property
    def sizeof_(self) -> str:
        return compute_module_size(
            self, requires_grad=False, include_buffers=True, human=True
        )

    @property
    def dtype(self) -> torch.dtype:
        try:
            return next(self.parameters()).dtype
        except StopIteration:
            return torch.float32
        except Exception as err:
            raise err  # unknown error

    @property
    def device(self) -> torch.device:
        try:
            return next(self.parameters()).device
        except StopIteration:
            return torch.device("cpu")
        except Exception as err:
            raise err  # unknown error

    @property
    def dtype_(self) -> str:
        return str(self.dtype).replace("torch.", "")

    @property
    def device_(self) -> str:
        return str(self.device)


class CkptMixin(object):
    """Mixin class for loading from checkpoint class methods"""

    @classmethod
    def from_checkpoint(
        cls, path: Union[str, Path], device: Optional[torch.device] = None
    ) -> Tuple[nn.Module, dict]:
        """Load a model from a checkpoint.

        Parameters
        ----------
        path : str or pathlib.Path
            Path of the checkpoint.
        device : torch.device, optional
            Map location of the model parameters,
            defaults to "cuda" if available, otherwise "cpu".

        Returns
        -------
        model : torch.nn.Module
            The model loaded from a checkpoint.
        aux_config : dict
            Auxiliary configs that are needed for data preprocessing, etc.

        """
        _device = device or DEFAULTS.device
        ckpt = torch.load(path, map_location=_device)
        aux_config = ckpt.get("train_config", None) or ckpt.get("config", None)
        assert (
            aux_config is not None
        ), "input checkpoint has no sufficient data to recover a model"
        kwargs = dict(
            config=ckpt["model_config"],
        )
        if "classes" in aux_config:
            kwargs["classes"] = aux_config["classes"]
        if "n_leads" in aux_config:
            kwargs["n_leads"] = aux_config["n_leads"]
        model = cls(**kwargs)
        model.load_state_dict(ckpt["model_state_dict"])
        return model, aux_config

    def save(self, path: Union[str, Path], train_config: CFG) -> None:
        """Save the model to disk.

        Parameters
        ----------
        path : str or pathlib.Path
            Path to save the model.
        train_config : CFG
            Config for training the model,
            used when one restores the model.

        Returns
        -------
        None

        """
        path = Path(path)
        if not path.parent.exists():
            path.parent.mkdir(parents=True)
        torch.save(
            {
                "model_state_dict": self.state_dict(),
                "model_config": self.config,
                "train_config": train_config,
            },
            path,
        )



"""
C(R)NN structure models, for classifying ECG arrhythmias, and other tasks.
"""

import warnings
from copy import deepcopy
from typing import Any, Optional, Sequence, Union, List

import numpy as np
import torch
from torch import Tensor, nn
from einops import rearrange
from einops.layers.torch import Rearrange

from ..cfg import CFG
from ..components.outputs import BaseOutput
from ..model_configs.ecg_crnn import ECG_CRNN_CONFIG
from ..utils.misc import CitationMixin
from ..utils.utils_nn import CkptMixin, SizeMixin
from ._nets import (
    GlobalContextBlock,
    NonLocalBlock,
    SEBlock,
    SelfAttention,
    MLP,
    StackedLSTM,
)
from .cnn.densenet import DenseNet
from .cnn.multi_scopic import MultiScopicCNN
from .cnn.resnet import ResNet
from .cnn.regnet import RegNet
from .cnn.mobilenet import MobileNetV1, MobileNetV2, MobileNetV3
from .cnn.vgg import VGG16
from .cnn.xception import Xception
from .transformers import Transformer


__all__ = [
    "ECG_CRNN",
]


class ECG_CRNN(nn.Module, CkptMixin, SizeMixin, CitationMixin):
    """Convolutional (Recurrent) Neural Network for ECG tasks.

    This C(R)NN architecture is adapted from [1]_, [2]_ in the first place,
    and then modified to be more general, and more flexible.
    The most famous model is perhaps [3]_, which is a modified 1D-ResNet34 model.
    The website of this model is [4]_, and the code is hosted on [5]_.
    The C(R)NN models have long been competitive in various ECG tasks,
    e.g. [6]_, [7]_. The models are also used in the PhysioNet/CinC Challenges.

    Parameters
    ----------
    classes : List[str]
        List of the names of the classes.
    n_leads : int
        Number of leads (number of input channels).
    config : dict
        Other hyper-parameters, including kernel sizes, etc.
        Refer to corresponding config files.

    References
    ----------
    .. [1] Yao, Qihang, et al.
           "Time-Incremental Convolutional Neural Network for Arrhythmia Detection in Varied-Length Electrocardiogram."
           2018 IEEE 16th Intl Conf on Dependable, Autonomic and Secure Computing,
           16th Intl Conf on Pervasive Intelligence and Computing,
           4th Intl Conf on Big Data Intelligence and Computing and Cyber Science and Technology Congress
           (DASC/PiCom/DataCom/CyberSciTech). IEEE, 2018.
    .. [2] Yao, Qihang, et al.
           "Multi-class Arrhythmia detection from 12-lead varied-length ECG using Attention-based
           Time-Incremental Convolutional Neural Network." Information Fusion 53 (2020): 174-182.
    .. [3] Hannun, Awni Y., et al.
           "Cardiologist-level arrhythmia detection and classification in ambulatory electrocardiograms
           using a deep neural network." Nature medicine 25.1 (2019): 65.
    .. [4] https://stanfordmlgroup.github.io/projects/ecg2/
    .. [5] https://github.com/awni/ecg
    .. [6] CPSC2018 entry 0236
    .. [7] CPSC2019 entry 0416

    """

    __name__ = "ECG_CRNN"

    def __init__(
        self,
        classes: Sequence[str],
        n_leads: int,
        config: Optional[CFG] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.classes = list(classes)
        self.n_classes = len(classes)
        self.n_leads = n_leads
        self.config = deepcopy(ECG_CRNN_CONFIG)
        if not config:
            warnings.warn(
                "No config is provided, using default config.", RuntimeWarning
            )
        self.config.update(deepcopy(config) or {})

        cnn_choice = self.config.cnn.name.lower()
        cnn_config = self.config.cnn[self.config.cnn.name]
        if "resnet" in cnn_choice or "resnext" in cnn_choice:
            self.cnn = ResNet(self.n_leads, **cnn_config)
        elif "regnet" in cnn_choice:
            self.cnn = RegNet(self.n_leads, **cnn_config)
        elif "multi_scopic" in cnn_choice:
            self.cnn = MultiScopicCNN(self.n_leads, **cnn_config)
        elif "mobile_net" in cnn_choice or "mobilenet" in cnn_choice:
            if "v1" in cnn_choice:
                self.cnn = MobileNetV1(self.n_leads, **cnn_config)
            elif "v2" in cnn_choice:
                self.cnn = MobileNetV2(self.n_leads, **cnn_config)
            elif "v3" in cnn_choice:
                self.cnn = MobileNetV3(self.n_leads, **cnn_config)
            else:
                raise ValueError(
                    f"CNN \042{cnn_choice}\042 is not supported for {self.__name__}"
                )
        elif "densenet" in cnn_choice or "dense_net" in cnn_choice:
            self.cnn = DenseNet(self.n_leads, **cnn_config)
        elif "vgg16" in cnn_choice:
            self.cnn = VGG16(self.n_leads, **cnn_config)
        elif "xception" in cnn_choice:
            self.cnn = Xception(self.n_leads, **cnn_config)
        else:
            raise NotImplementedError(f"CNN \042{cnn_choice}\042 not implemented yet")
        rnn_input_size = self.cnn.compute_output_shape(None, None)[1]

        if self.config.rnn.name.lower() == "none":
            self.rnn_in_rearrange = Rearrange(
                "batch_size channels seq_len -> seq_len batch_size channels"
            )
            self.rnn = nn.Identity()
            self.__rnn_seqlen_dim = 0
            self.rnn_out_rearrange = nn.Identity()
            attn_input_size = rnn_input_size
        elif self.config.rnn.name.lower() == "lstm":
            self.rnn_in_rearrange = Rearrange(
                "batch_size channels seq_len -> seq_len batch_size channels"
            )
            self.rnn = StackedLSTM(
                input_size=rnn_input_size,
                hidden_sizes=self.config.rnn.lstm.hidden_sizes,
                bias=self.config.rnn.lstm.bias,
                dropouts=self.config.rnn.lstm.dropouts,
                bidirectional=self.config.rnn.lstm.bidirectional,
                return_sequences=self.config.rnn.lstm.retseq,
            )
            self.__rnn_seqlen_dim = 0
            self.rnn_out_rearrange = nn.Identity()
            attn_input_size = self.rnn.compute_output_shape(None, None)[-1]
        elif self.config.rnn.name.lower() == "linear":
            # abuse of notation, to put before the global attention module
            self.rnn_in_rearrange = Rearrange(
                "batch_size channels seq_len -> batch_size seq_len channels"
            )
            self.rnn = MLP(
                in_channels=rnn_input_size,
                out_channels=self.config.rnn.linear.out_channels,
                activation=self.config.rnn.linear.activation,
                bias=self.config.rnn.linear.bias,
                dropouts=self.config.rnn.linear.dropouts,
            )
            self.__rnn_seqlen_dim = 1
            self.rnn_out_rearrange = Rearrange(
                "batch_size seq_len channels -> seq_len batch_size channels"
            )
            attn_input_size = self.rnn.compute_output_shape(None, None)[-1]
        else:
            raise NotImplementedError(
                f"RNN \042{self.config.rnn.name}\042 not implemented yet"
            )

        # attention
        if self.config.rnn.name.lower() == "lstm" and not self.config.rnn.lstm.retseq:
            self.attn_in_rearrange = nn.Identity()
            self.attn = nn.Identity()
            self.__attn_seqlen_dim = 0
            self.attn_out_rearrange = nn.Identity()
            clf_input_size = attn_input_size
            if self.config.attn.name.lower() != "none":
                warnings.warn(
                    f"since `retseq` of rnn is False, hence attention `{self.config.attn.name}` is ignored",
                    RuntimeWarning,
                )
        elif self.config.attn.name.lower() == "none":
            self.attn_in_rearrange = Rearrange(
                "seq_len batch_size channels -> batch_size channels seq_len"
            )
            self.attn = nn.Identity()
            self.__attn_seqlen_dim = -1
            self.attn_out_rearrange = nn.Identity()
            clf_input_size = attn_input_size
        elif self.config.attn.name.lower() == "nl":  # non_local
            self.attn_in_rearrange = Rearrange(
                "seq_len batch_size channels -> batch_size channels seq_len"
            )
            self.attn = NonLocalBlock(
                in_channels=attn_input_size,
                filter_lengths=self.config.attn.nl.filter_lengths,
                subsample_length=self.config.attn.nl.subsample_length,
                batch_norm=self.config.attn.nl.batch_norm,
            )
            self.__attn_seqlen_dim = -1
            self.attn_out_rearrange = nn.Identity()
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "se":  # squeeze_exitation
            self.attn_in_rearrange = Rearrange(
                "seq_len batch_size channels -> batch_size channels seq_len"
            )
            self.attn = SEBlock(
                in_channels=attn_input_size,
                reduction=self.config.attn.se.reduction,
                activation=self.config.attn.se.activation,
                kw_activation=self.config.attn.se.kw_activation,
                bias=self.config.attn.se.bias,
            )
            self.__attn_seqlen_dim = -1
            self.attn_out_rearrange = nn.Identity()
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "gc":  # global_context
            self.attn_in_rearrange = Rearrange(
                "seq_len batch_size channels -> batch_size channels seq_len"
            )
            self.attn = GlobalContextBlock(
                in_channels=attn_input_size,
                ratio=self.config.attn.gc.ratio,
                reduction=self.config.attn.gc.reduction,
                pooling_type=self.config.attn.gc.pooling_type,
                fusion_types=self.config.attn.gc.fusion_types,
            )
            self.__attn_seqlen_dim = -1
            self.attn_out_rearrange = nn.Identity()
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "sa":  # self_attention
            # NOTE: this branch NOT tested
            self.attn_in_rearrange = nn.Identity()
            self.attn = SelfAttention(
                embed_dim=attn_input_size,
                num_heads=self.config.attn.sa.get(
                    "num_heads", self.config.attn.sa.get("head_num")
                ),
                dropout=self.config.attn.sa.dropout,
                bias=self.config.attn.sa.bias,
            )
            self.__attn_seqlen_dim = 0
            self.attn_out_rearrange = Rearrange(
                "seq_len batch_size channels -> batch_size channels seq_len"
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[-1]
        elif self.config.attn.name.lower() == "transformer":
            self.attn = Transformer(
                input_size=attn_input_size,
                hidden_size=self.config.attn.transformer.hidden_size,
                num_layers=self.config.attn.transformer.num_layers,
                num_heads=self.config.attn.transformer.num_heads,
                dropout=self.config.attn.transformer.dropout,
                activation=self.config.attn.transformer.activation,
            )
            if self.attn.batch_first:
                self.attn_in_rearrange = Rearrange(
                    "seq_len batch_size channels -> batch_size seq_len channels"
                )
                self.attn_out_rearrange = Rearrange(
                    "batch_size seq_len channels -> batch_size channels seq_len"
                )
                self.__attn_seqlen_dim = 1
            else:
                self.attn_in_rearrange = nn.Identity()
                self.attn_out_rearrange = Rearrange(
                    "seq_len batch_size channels -> batch_size channels seq_len"
                )
                self.__attn_seqlen_dim = 0
            clf_input_size = self.attn.compute_output_shape(None, None)[-1]
        else:
            raise NotImplementedError(
                f"Attention \042{self.config.attn.name}\042 not implemented yet"
            )

        # global pooling
        if self.config.rnn.name.lower() == "lstm" and not self.config.rnn.lstm.retseq:
            self.pool = nn.Identity()
            if self.config.global_pool.lower() != "none":
                warnings.warn(
                    f"since `retseq` of rnn is False, hence global pooling `{self.config.global_pool}` is ignored",
                    RuntimeWarning,
                )
            self.pool_rearrange = nn.Identity()
            self.__clf_input_seq = False
        elif self.config.global_pool.lower() == "max":
            self.pool = nn.AdaptiveMaxPool1d(
                (self.config.global_pool_size,), return_indices=False
            )
            clf_input_size *= self.config.global_pool_size
            self.pool_rearrange = Rearrange(
                "batch_size channels pool_size -> batch_size (channels pool_size)"
            )
            self.__clf_input_seq = False
        elif self.config.global_pool.lower() == "avg":
            self.pool = nn.AdaptiveAvgPool1d((self.config.global_pool_size,))
            clf_input_size *= self.config.global_pool_size
            self.pool_rearrange = Rearrange(
                "batch_size channels pool_size -> batch_size (channels pool_size)"
            )
            self.__clf_input_seq = False
        elif self.config.global_pool.lower() == "attn":
            raise NotImplementedError("Attentive pooling not implemented yet!")
        elif self.config.global_pool.lower() == "none":
            self.pool = nn.Identity()
            self.pool_rearrange = Rearrange(
                "batch_size channels seq_len -> batch_size seq_len channels"
            )
            self.__clf_input_seq = True
        else:
            raise NotImplementedError(
                f"Global Pooling \042{self.config.global_pool}\042 not implemented yet!"
            )

        # input of `self.clf` has shape: batch_size, channels
        self.clf = MLP(
            in_channels=clf_input_size,
            out_channels=self.config.clf.out_channels + [self.n_classes],
            activation=self.config.clf.activation,
            bias=self.config.clf.bias,
            dropouts=self.config.clf.dropouts,
            skip_last_activation=True,
        )

        # for inference
        # classification: if single-label, use softmax; otherwise (multi-label) use sigmoid
        # sequence tagging: if background counted in `classes`, use softmax; otherwise use sigmoid
        self.sigmoid = nn.Sigmoid()
        self.softmax = nn.Softmax(-1)

    def extract_features(self, input: Tensor) -> Tensor:
        """Extract feature map before the
        dense (linear) classifying layer(s).

        Parameters
        ----------
        input : torch.Tensor
            Input signal tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        features : torch.Tensor
            Feature map tensor,
            of shape ``(batch_size, channels, seq_len)``
            or ``(batch_size, channels)``.

        """
        # CNN
        features = self.cnn(input)  # batch_size, channels, seq_len

        # RNN (optional)
        features = self.rnn_in_rearrange(features)
        features = self.rnn(features)
        features = self.rnn_out_rearrange(features)

        # Attention (optional)
        features = self.attn_in_rearrange(features)
        features = self.attn(features)
        features = self.attn_out_rearrange(features)

        return features

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the model.

        Parameters
        ----------
        input : torch.Tensor
            Input signal tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        pred : torch.Tensor
            Predictions tensor,
            of shape ``(batch_size, seq_len, channels)``
            or ``(batch_size, channels)``.

        """
        features = self.extract_features(input)

        # global pooling (optional)
        features = self.pool(features)
        features = self.pool_rearrange(features)

        pred = self.clf(features)

        return pred

    @torch.no_grad()
    def inference(
        self,
        input: Union[np.ndarray, Tensor],
        class_names: bool = False,
        bin_pred_thr: float = 0.5,
    ) -> BaseOutput:
        """Inference method for the model.

        Parameters
        ----------
        input : numpy.ndarray or torch.Tensor
            Input tensor, of shape ``(batch_size, channels, seq_len)``.
        class_names : bool, default False
            If True, the returned scalar predictions will be
            a :class:`~pandas.DataFrame`,
            with class names for each scalar prediction.
        bin_pred_thr : float, default 0.5
            Threshold for making binary predictions from scalar predictions.

        Returns
        -------
        output : BaseOutput
            The output of the inference method, including the following items:

                - prob: numpy.ndarray or torch.Tensor,
                  scalar predictions, (and binary predictions if `class_names` is True).
                - pred: numpy.ndarray or torch.Tensor,
                  the array (with values 0, 1 for each class) of binary prediction.

        """
        raise NotImplementedError("implement a task specific inference method")

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input signal tensor.
        batch_size : int, optional
            Batch size of the input signal tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the model.

        """
        output_shape = self.cnn.compute_output_shape(seq_len, batch_size)
        _, _, _seq_len = output_shape
        if self.rnn.__class__.__name__ != "Identity":
            output_shape = self.rnn.compute_output_shape(_seq_len, batch_size)
            _seq_len = output_shape[self.__rnn_seqlen_dim]
        if self.attn.__class__.__name__ != "Identity":
            output_shape = self.attn.compute_output_shape(_seq_len, batch_size)
            _seq_len = output_shape[self.__attn_seqlen_dim]
        if self.clf.__class__.__name__ != "Identity":
            output_shape = self.clf.compute_output_shape(
                _seq_len, batch_size, input_seq=self.__clf_input_seq
            )
        return output_shape

    @property
    def doi(self) -> List[str]:
        doi = []
        candidates = [self.config]
        while len(candidates) > 0:
            new_candidates = []
            for candidate in candidates:
                if hasattr(candidate, "doi"):
                    if isinstance(candidate.doi, str):
                        doi.append(candidate.doi)
                    else:
                        doi.extend(list(candidate.doi))
                for k, v in candidate.items():
                    if isinstance(v, CFG):
                        new_candidates.append(v)
            candidates = new_candidates
        doi = list(
            set(doi + ["10.1016/j.inffus.2019.06.024", "10.1088/1361-6579/ac6aa3"])
        )
        return doi

    @classmethod
    def from_v1(cls, v1_ckpt: str, device: Optional[torch.device] = None) -> "ECG_CRNN":
        """Restore an instance of the model from a v1 checkpoint.

        Parameters
        ----------
        v1_ckpt : str
            Path to the v1 checkpoint file.

        Returns
        -------
        model : ECG_CRNN
            The model instance restored from the v1 checkpoint.

        """
        v1_model, _ = ECG_CRNN_v1.from_checkpoint(v1_ckpt, device=device)
        model = cls(
            classes=v1_model.classes, n_leads=v1_model.n_leads, config=v1_model.config
        )
        model = model.to(v1_model.device)
        model.cnn.load_state_dict(v1_model.cnn.state_dict())
        if model.rnn.__class__.__name__ != "Identity":
            model.rnn.load_state_dict(v1_model.rnn.state_dict())
        if model.attn.__class__.__name__ != "Identity":
            model.attn.load_state_dict(v1_model.attn.state_dict())
        model.clf.load_state_dict(v1_model.clf.state_dict())
        del v1_model
        return model


class ECG_CRNN_v1(nn.Module, CkptMixin, SizeMixin, CitationMixin):
    """Convolutional (Recurrent) Neural Network for ECG tasks.

    This C(R)NN architecture is adapted from [1]_, [2]_ in the first place,
    and then modified to be more general, and more flexible.
    The most famous model is perhaps [3]_, which is a modified 1D-ResNet34 model.
    The website of this model is [4]_, and the code is hosted on [5]_.
    The C(R)NN models have long been competitive in various ECG tasks,
    e.g. [6]_, [7]_. The models are also used in the PhysioNet/CinC Challenges.

    Parameters
    ----------
    classes : List[str]
        List of the names of the classes.
    n_leads : int
        Number of leads (number of input channels).
    config : dict
        Other hyper-parameters, including kernel sizes, etc.
        Refer to corresponding config files.

    References
    ----------
    .. [1] Yao, Qihang, et al.
           "Time-Incremental Convolutional Neural Network for Arrhythmia Detection in Varied-Length Electrocardiogram."
           2018 IEEE 16th Intl Conf on Dependable, Autonomic and Secure Computing,
           16th Intl Conf on Pervasive Intelligence and Computing,
           4th Intl Conf on Big Data Intelligence and Computing and Cyber Science and Technology Congress
           (DASC/PiCom/DataCom/CyberSciTech). IEEE, 2018.
    .. [2] Yao, Qihang, et al.
           "Multi-class Arrhythmia detection from 12-lead varied-length ECG using Attention-based
           Time-Incremental Convolutional Neural Network." Information Fusion 53 (2020): 174-182.
    .. [3] Hannun, Awni Y., et al.
           "Cardiologist-level arrhythmia detection and classification in ambulatory electrocardiograms
           using a deep neural network." Nature medicine 25.1 (2019): 65.
    .. [4] https://stanfordmlgroup.github.io/projects/ecg2/
    .. [5] https://github.com/awni/ecg
    .. [6] CPSC2018 entry 0236
    .. [7] CPSC2019 entry 0416

    """

    __name__ = "ECG_CRNN_v1"

    def __init__(
        self,
        classes: Sequence[str],
        n_leads: int,
        config: Optional[CFG] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.classes = list(classes)
        self.n_classes = len(classes)
        self.n_leads = n_leads
        self.config = deepcopy(ECG_CRNN_CONFIG)
        if not config:
            warnings.warn(
                "No config is provided, using default config.", RuntimeWarning
            )
        self.config.update(deepcopy(config) or {})

        cnn_choice = self.config.cnn.name.lower()
        cnn_config = self.config.cnn[self.config.cnn.name]
        if "resnet" in cnn_choice or "resnext" in cnn_choice:
            self.cnn = ResNet(self.n_leads, **cnn_config)
        elif "regnet" in cnn_choice:
            self.cnn = RegNet(self.n_leads, **cnn_config)
        elif "multi_scopic" in cnn_choice:
            self.cnn = MultiScopicCNN(self.n_leads, **cnn_config)
        elif "mobile_net" in cnn_choice or "mobilenet" in cnn_choice:
            if "v1" in cnn_choice:
                self.cnn = MobileNetV1(self.n_leads, **cnn_config)
            elif "v2" in cnn_choice:
                self.cnn = MobileNetV2(self.n_leads, **cnn_config)
            elif "v3" in cnn_choice:
                self.cnn = MobileNetV3(self.n_leads, **cnn_config)
            else:
                raise ValueError(
                    f"CNN \042{cnn_choice}\042 is not supported for {self.__name__}"
                )
        elif "densenet" in cnn_choice or "dense_net" in cnn_choice:
            self.cnn = DenseNet(self.n_leads, **cnn_config)
        elif "vgg16" in cnn_choice:
            self.cnn = VGG16(self.n_leads, **cnn_config)
        elif "xception" in cnn_choice:
            self.cnn = Xception(self.n_leads, **cnn_config)
        else:
            raise NotImplementedError(f"CNN \042{cnn_choice}\042 not implemented yet")
        rnn_input_size = self.cnn.compute_output_shape(None, None)[1]

        if self.config.rnn.name.lower() == "none":
            self.rnn = None
            attn_input_size = rnn_input_size
        elif self.config.rnn.name.lower() == "lstm":
            self.rnn = StackedLSTM(
                input_size=rnn_input_size,
                hidden_sizes=self.config.rnn.lstm.hidden_sizes,
                bias=self.config.rnn.lstm.bias,
                dropouts=self.config.rnn.lstm.dropouts,
                bidirectional=self.config.rnn.lstm.bidirectional,
                return_sequences=self.config.rnn.lstm.retseq,
            )
            attn_input_size = self.rnn.compute_output_shape(None, None)[-1]
        elif self.config.rnn.name.lower() == "linear":
            # abuse of notation, to put before the global attention module
            self.rnn = MLP(
                in_channels=rnn_input_size,
                out_channels=self.config.rnn.linear.out_channels,
                activation=self.config.rnn.linear.activation,
                bias=self.config.rnn.linear.bias,
                dropouts=self.config.rnn.linear.dropouts,
            )
            attn_input_size = self.rnn.compute_output_shape(None, None)[-1]
        else:
            raise NotImplementedError(
                f"RNN \042{self.config.rnn.name}\042 not implemented yet"
            )

        # attention
        if self.config.rnn.name.lower() == "lstm" and not self.config.rnn.lstm.retseq:
            self.attn = None
            clf_input_size = attn_input_size
            if self.config.attn.name.lower() != "none":
                warnings.warn(
                    f"since `retseq` of rnn is False, hence attention `{self.config.attn.name}` is ignored",
                    RuntimeWarning,
                )
        elif self.config.attn.name.lower() == "none":
            self.attn = None
            clf_input_size = attn_input_size
        elif self.config.attn.name.lower() == "nl":  # non_local
            self.attn = NonLocalBlock(
                in_channels=attn_input_size,
                filter_lengths=self.config.attn.nl.filter_lengths,
                subsample_length=self.config.attn.nl.subsample_length,
                batch_norm=self.config.attn.nl.batch_norm,
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "se":  # squeeze_exitation
            self.attn = SEBlock(
                in_channels=attn_input_size,
                reduction=self.config.attn.se.reduction,
                activation=self.config.attn.se.activation,
                kw_activation=self.config.attn.se.kw_activation,
                bias=self.config.attn.se.bias,
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "gc":  # global_context
            self.attn = GlobalContextBlock(
                in_channels=attn_input_size,
                ratio=self.config.attn.gc.ratio,
                reduction=self.config.attn.gc.reduction,
                pooling_type=self.config.attn.gc.pooling_type,
                fusion_types=self.config.attn.gc.fusion_types,
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "sa":  # self_attention
            # NOTE: this branch NOT tested
            self.attn = SelfAttention(
                embed_dim=attn_input_size,
                num_heads=self.config.attn.sa.get(
                    "num_heads", self.config.attn.sa.get("head_num")
                ),
                dropout=self.config.attn.sa.dropout,
                bias=self.config.attn.sa.bias,
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[-1]
        elif self.config.attn.name.lower() == "transformer":
            self.attn = Transformer(
                input_size=attn_input_size,
                hidden_size=self.config.attn.transformer.hidden_size,
                num_layers=self.config.attn.transformer.num_layers,
                num_heads=self.config.attn.transformer.num_heads,
                dropout=self.config.attn.transformer.dropout,
                activation=self.config.attn.transformer.activation,
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[-1]
        else:
            raise NotImplementedError(
                f"Attention \042{self.config.attn.name}\042 not implemented yet"
            )

        if self.config.rnn.name.lower() == "lstm" and not self.config.rnn.lstm.retseq:
            self.pool = None
            if self.config.global_pool.lower() != "none":
                warnings.warn(
                    f"since `retseq` of rnn is False, hence global pooling `{self.config.global_pool}` is ignored",
                    RuntimeWarning,
                )
        elif self.config.global_pool.lower() == "max":
            self.pool = nn.AdaptiveMaxPool1d(
                (self.config.global_pool_size,), return_indices=False
            )
            clf_input_size *= self.config.global_pool_size
        elif self.config.global_pool.lower() == "avg":
            self.pool = nn.AdaptiveAvgPool1d((self.config.global_pool_size,))
            clf_input_size *= self.config.global_pool_size
        elif self.config.global_pool.lower() == "attn":
            raise NotImplementedError("Attentive pooling not implemented yet!")
        elif self.config.global_pool.lower() == "none":
            self.pool = None
        else:
            raise NotImplementedError(
                f"Global Pooling \042{self.config.global_pool}\042 not implemented yet!"
            )

        # input of `self.clf` has shape: batch_size, channels
        self.clf = MLP(
            in_channels=clf_input_size,
            out_channels=self.config.clf.out_channels + [self.n_classes],
            activation=self.config.clf.activation,
            bias=self.config.clf.bias,
            dropouts=self.config.clf.dropouts,
            skip_last_activation=True,
        )

        # for inference
        # classification: if single-label, use softmax; otherwise (multi-label) use sigmoid
        # sequence tagging: if background counted in `classes`, use softmax; otherwise use sigmoid
        self.sigmoid = nn.Sigmoid()
        self.softmax = nn.Softmax(-1)

    def extract_features(self, input: Tensor) -> Tensor:
        """Extract feature map before the
        dense (linear) classifying layer(s).

        Parameters
        ----------
        input : torch.Tensor
            Input signal tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        features : torch.Tensor
            Feature map tensor,
            of shape ``(batch_size, channels, seq_len)``
            or ``(batch_size, channels)``.

        """
        # CNN
        features = self.cnn(input)  # batch_size, channels, seq_len

        # RNN (optional)
        if self.config.rnn.name.lower() in ["lstm"]:
            # (batch_size, channels, seq_len) --> (seq_len, batch_size, channels)
            features = features.permute(2, 0, 1)
            features = self.rnn(
                features
            )  # (seq_len, batch_size, channels) or (batch_size, channels)
        elif self.config.rnn.name.lower() in ["linear"]:
            # (batch_size, channels, seq_len) --> (batch_size, seq_len, channels)
            features = features.permute(0, 2, 1)
            features = self.rnn(features)  # (batch_size, seq_len, channels)
            # (batch_size, seq_len, channels) --> (seq_len, batch_size, channels)
            features = features.permute(1, 0, 2)
        else:
            # (batch_size, channels, seq_len) --> (seq_len, batch_size, channels)
            features = features.permute(2, 0, 1)

        # Attention (optional)
        if self.attn is None and features.ndim == 3:
            # (seq_len, batch_size, channels) --> (batch_size, channels, seq_len)
            features = features.permute(1, 2, 0)
        elif self.config.attn.name.lower() in ["nl", "se", "gc"]:
            # (seq_len, batch_size, channels) --> (batch_size, channels, seq_len)
            features = features.permute(1, 2, 0)
            features = self.attn(features)  # (batch_size, channels, seq_len)
        elif self.config.attn.name.lower() in ["sa"]:
            features = self.attn(features)  # (seq_len, batch_size, channels)
            # (seq_len, batch_size, channels) -> (batch_size, channels, seq_len)
            features = features.permute(1, 2, 0)
        elif self.config.attn.name.lower() in ["transformer"]:
            features = self.attn(features)
            # (seq_len, batch_size, channels) -> (batch_size, channels, seq_len)
            features = features.permute(1, 2, 0)
        return features

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the model.

        Parameters
        ----------
        input : torch.Tensor
            Input signal tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        pred : torch.Tensor
            Predictions tensor,
            of shape ``(batch_size, seq_len, channels)``
            or ``(batch_size, channels)``.

        """
        features = self.extract_features(input)

        if self.pool:
            features = self.pool(features)  # (batch_size, channels, pool_size)
            # features = features.squeeze(dim=-1)
            features = rearrange(
                features,
                "batch_size channels pool_size -> batch_size (channels pool_size)",
            )
        elif features.ndim == 3:
            # (batch_size, channels, seq_len) --> (batch_size, seq_len, channels)
            features = features.permute(0, 2, 1)

        # print(f"clf in shape = {features.shape}")
        pred = self.clf(features)  # batch_size, n_classes

        return pred

    @torch.no_grad()
    def inference(
        self,
        input: Union[np.ndarray, Tensor],
        class_names: bool = False,
        bin_pred_thr: float = 0.5,
    ) -> BaseOutput:
        """Inference method for the model.

        Parameters
        ----------
        input : numpy.ndarray or torch.Tensor
            Input tensor, of shape ``(batch_size, channels, seq_len)``.
        class_names : bool, default False
            If True, the returned scalar predictions will be
            a :class:`~pandas.DataFrame`,
            with class names for each scalar prediction.
        bin_pred_thr : float, default 0.5
            Threshold for making binary predictions from scalar predictions.

        Returns
        -------
        output : BaseOutput
            The output of the inference method, including the following items:

                - prob: numpy.ndarray or torch.Tensor,
                  scalar predictions, (and binary predictions if `class_names` is True).
                - pred: numpy.ndarray or torch.Tensor,
                  the array (with values 0, 1 for each class) of binary prediction.

        """
        raise NotImplementedError("implement a task specific inference method")

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input signal tensor.
        batch_size : int, optional
            Batch size of the input signal tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the model.

        """
        if self.pool:
            return (batch_size, len(self.classes))
        else:
            _seq_len = seq_len
            output_shape = self.cnn.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
            if self.rnn:
                output_shape = self.rnn.compute_output_shape(_seq_len, batch_size)
                _seq_len = output_shape[0]
            if self.attn:
                output_shape = self.attn.compute_output_shape(_seq_len, batch_size)
                _seq_len = output_shape[-1]
            output_shape = self.clf.compute_output_shape(_seq_len, batch_size)
            return output_shape

    @property
    def doi(self) -> List[str]:
        doi = []
        candidates = [self.config]
        while len(candidates) > 0:
            new_candidates = []
            for candidate in candidates:
                if hasattr(candidate, "doi"):
                    if isinstance(candidate.doi, str):
                        doi.append(candidate.doi)
                    else:
                        doi.extend(list(candidate.doi))
                for k, v in candidate.items():
                    if isinstance(v, CFG):
                        new_candidates.append(v)
            candidates = new_candidates
        doi = list(
            set(doi + ["10.1016/j.inffus.2019.06.024", "10.1088/1361-6579/ac6aa3"])
        )
        return doi



"""
torch_ecg.databases
===================

This module contains ECG database readers.

.. contents:: torch_ecg.databases
    :depth: 2
    :local:
    :backlinks: top

.. currentmodule:: torch_ecg.databases

Base classes
------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    PhysioNetDataBase
    NSRRDataBase
    CPSCDataBase
    PSGDataBaseMixin

PhysioNet database readers
--------------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    AFDB
    ApneaECG
    CINC2017
    CINC2018
    CINC2020
    CINC2021
    LTAFDB
    LUDB
    MITDB
    QTDB

CPSC database readers
---------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    CPSC2018
    CPSC2019
    CPSC2020
    CPSC2021

NSRR database readers
---------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    SHHS

Other database readers
----------------------
.. autosummary::
    :toctree: generated/
    :recursive:

    CACHET_CADB
    SPH

Miscellaneous
-------------
.. autosummary::
    :toctree: generated/
    :recursive:

    BeatAnn

"""

from . import aux_data
from .base import (
    PhysioNetDataBase,
    NSRRDataBase,
    CPSCDataBase,
    DataBaseInfo,
    BeatAnn,
    WFDB_Beat_Annotations,
    WFDB_Non_Beat_Annotations,
    WFDB_Rhythm_Annotations,
    PSGDataBaseMixin,
)
from .cpsc_databases import CPSC2018, CPSC2019, CPSC2020, CPSC2021
from .nsrr_databases import SHHS
from .other_databases import CACHET_CADB, SPH
from .physionet_databases import (
    AFDB,
    ApneaECG,
    CINC2017,
    CINC2018,
    CINC2020,
    CINC2021,
    LTAFDB,
    LUDB,
    MITDB,
    QTDB,
)


__all__ = [
    # from physionet
    "AFDB",
    "ApneaECG",
    "CINC2017",
    "CINC2018",
    "CINC2020",
    "CINC2021",
    "LTAFDB",
    "LUDB",
    "MITDB",
    "QTDB",
    # from CPSC
    "CPSC2018",
    "CPSC2019",
    "CPSC2020",
    "CPSC2021",
    # from NSRR
    "SHHS",
    # other databases
    "CACHET_CADB",
    "SPH",
    # base classes
    "PhysioNetDataBase",
    "NSRRDataBase",
    "CPSCDataBase",
    "PSGDataBaseMixin",
    # auxilliary data, functions and classes
    "aux_data",
    "WFDB_Beat_Annotations",
    "WFDB_Non_Beat_Annotations",
    "WFDB_Rhythm_Annotations",
    "BeatAnn",
    "DataBaseInfo",
    "list_databases",
]


def list_databases() -> list:
    return [
        db
        for db in __all__
        if db
        not in [
            "PhysioNetDataBase",
            "NSRRDataBase",
            "CPSCDataBase",
            "PSGDataBaseMixin",
            "aux_data",
            "WFDB_Beat_Annotations",
            "WFDB_Non_Beat_Annotations",
            "WFDB_Rhythm_Annotations",
            "BeatAnn",
            "DataBaseInfo",
            "list_databases",
        ]
    ]



"""
Designing Network Design Spaces
"""

from copy import deepcopy

from ...cfg import CFG
from .resnet import (
    resnet_stem,
    resnet_stem_C,
    resnet_bottle_neck,
    resnet_bottle_neck_B,
    resnet_bottle_neck_D,
    resnet_bottle_neck_se,
    resnet_bottle_neck_nl,
    resnet_bottle_neck_gc,
)


__all__ = [
    "regnet_16_8",
    "regnet_27_24",
    "regnet_23_168",
    "regnet_S",
    "regnet_bottle_neck",
    "regnet_bottle_neck_B",
    "regnet_bottle_neck_D",
    "regnet_bottle_neck_se",
    "regnet_bottle_neck_nl",
    "regnet_bottle_neck_gc",
]


regnet_stem = deepcopy(resnet_stem)
regnet_stem_C = deepcopy(resnet_stem_C)
regnet_bottle_neck = deepcopy(resnet_bottle_neck)
regnet_bottle_neck_B = deepcopy(resnet_bottle_neck_B)
regnet_bottle_neck_D = deepcopy(resnet_bottle_neck_D)
regnet_bottle_neck_se = deepcopy(resnet_bottle_neck_se)
regnet_bottle_neck_nl = deepcopy(resnet_bottle_neck_nl)
regnet_bottle_neck_gc = deepcopy(resnet_bottle_neck_gc)


regnet_16_8 = CFG()
regnet_16_8.fs = 500
regnet_16_8.stem = deepcopy(resnet_stem)
regnet_16_8.filter_lengths = 11
regnet_16_8.subsample_lengths = 2

regnet_16_8.tot_blocks = 16
regnet_16_8.group_widths = 8
regnet_16_8.w_a = 27.89
regnet_16_8.w_0 = 48
regnet_16_8.w_m = 2.09

regnet_16_8.block = deepcopy(resnet_bottle_neck_se)
regnet_16_8.block.expansion = 1


regnet_27_24 = deepcopy(regnet_16_8)
regnet_27_24.tot_blocks = 27
regnet_27_24.group_widths = 24
regnet_27_24.w_a = 20.71
regnet_27_24.w_0 = 48
regnet_27_24.w_m = 2.65


regnet_23_168 = deepcopy(regnet_16_8)
regnet_23_168.tot_blocks = 23
regnet_23_168.group_widths = 168
regnet_23_168.w_a = 69.86
regnet_23_168.w_0 = 320
regnet_23_168.w_m = 2.0


regnet_S = CFG()
regnet_S.fs = 500
regnet_S.stem = deepcopy(resnet_stem)

regnet_S.num_blocks = [2, 2, 2, 2]
regnet_S.filter_lengths = [13, 11, 7, 5]
regnet_S.subsample_lengths = [2, 2, 2, 2]
regnet_S.num_filters = [32, 64, 128, 256]
regnet_S.group_widths = [8, 16, 32, 64]
regnet_S.block = deepcopy(resnet_bottle_neck_se)
regnet_S.block.expansion = 4



"""
"""

from itertools import repeat

from ...cfg import CFG

__all__ = [
    # vanilla xception
    "xception_vanilla",
    # custom xception
    "xception_leadwise",
]


xception_vanilla = CFG()
xception_vanilla.fs = 500
xception_vanilla.groups = 1
_base_num_filters = 8
xception_vanilla.entry_flow = CFG(
    init_num_filters=[_base_num_filters * 4, _base_num_filters * 8],
    init_filter_lengths=31,
    init_subsample_lengths=[2, 1],
    num_filters=[
        _base_num_filters * 16,
        _base_num_filters * 32,
        _base_num_filters * 91,
    ],
    filter_lengths=15,
    subsample_lengths=2,
    subsample_kernels=3,
)
xception_vanilla.middle_flow = CFG(
    num_filters=list(repeat(_base_num_filters * 91, 8)),
    filter_lengths=13,
)
xception_vanilla.exit_flow = CFG(
    final_num_filters=[_base_num_filters * 182, _base_num_filters * 256],
    final_filter_lengths=3,
    num_filters=[[_base_num_filters * 91, _base_num_filters * 128]],
    filter_lengths=17,
    subsample_lengths=2,
    subsample_kernels=3,
)

xception_leadwise = CFG()
xception_leadwise.fs = 500
xception_leadwise.groups = 12
_base_num_filters = 12 * 2
xception_leadwise.entry_flow = CFG(
    init_num_filters=[_base_num_filters * 4, _base_num_filters * 8],
    init_filter_lengths=31,
    init_subsample_lengths=[2, 1],
    num_filters=[
        _base_num_filters * 16,
        _base_num_filters * 32,
        _base_num_filters * 91,
    ],
    filter_lengths=15,
    subsample_lengths=2,
    subsample_kernels=3,
)
xception_leadwise.middle_flow = CFG(
    num_filters=list(repeat(_base_num_filters * 91, 8)),
    filter_lengths=13,
)
xception_leadwise.exit_flow = CFG(
    final_num_filters=[_base_num_filters * 182, _base_num_filters * 256],
    final_filter_lengths=17,
    num_filters=[[_base_num_filters * 91, _base_num_filters * 128]],
    filter_lengths=3,
    subsample_lengths=2,
    subsample_kernels=3,
)



"""
References
----------
[1] Moskalenko, Viktor, Nikolai Zolotykh, and Grigory Osipov. "Deep Learning for ECG Segmentation." International Conference on Neuroinformatics. Springer, Cham, 2019.
"""

from copy import deepcopy

from ....cfg import CFG, DEFAULTS
from ....utils import ecg_arrhythmia_knowledge as EAK

__all__ = [
    "LUDBTrainCfg",
]


_NAME = "ludb"


LUDBTrainCfg = CFG()
LUDBTrainCfg.fs = 500  # Hz, LUDB data fs
LUDBTrainCfg.classes = [
    "p",  # pwave
    "N",  # qrs complex
    "t",  # twave
    "i",  # isoelectric
]
# LUDBTrainCfg.mask_classes = [
#     "p",  # pwave
#     "N",  # qrs complex
#     "t",  # twave
# ]
LUDBTrainCfg.mask_classes = deepcopy(LUDBTrainCfg.classes)
LUDBTrainCfg.class_map = CFG(p=1, N=2, t=3, i=0)
# LUDBTrainCfg.mask_class_map = CFG({k:v-1 for k,v in LUDBTrainCfg.class_map.items() if k!="i"})
LUDBTrainCfg.mask_class_map = deepcopy(LUDBTrainCfg.class_map)

LUDBTrainCfg.db_dir = None
LUDBTrainCfg.log_dir = DEFAULTS.log_dir / _NAME
LUDBTrainCfg.model_dir = DEFAULTS.model_dir / _NAME
LUDBTrainCfg.checkpoints = DEFAULTS.checkpoints / _NAME
LUDBTrainCfg.log_dir.mkdir(parents=True, exist_ok=True)
LUDBTrainCfg.model_dir.mkdir(parents=True, exist_ok=True)
LUDBTrainCfg.checkpoints.mkdir(parents=True, exist_ok=True)
LUDBTrainCfg.keep_checkpoint_max = 20

LUDBTrainCfg.bias_thr = int(0.075 * LUDBTrainCfg.fs)  # TODO: renew this const
# detected waves that are within `skip_dist` from two ends of the signal will be ignored,
LUDBTrainCfg.skip_dist = int(0.5 * LUDBTrainCfg.fs)
LUDBTrainCfg.torch_dtype = DEFAULTS.DTYPE.TORCH
LUDBTrainCfg.np_dtype = DEFAULTS.DTYPE.NP

LUDBTrainCfg.fs = 500
LUDBTrainCfg.train_ratio = 0.8

LUDBTrainCfg.leads = (
    EAK.Standard12Leads
)  # ["II",]  # the lead to tain model, None --> all leads
LUDBTrainCfg.use_single_lead = (
    False  # use single lead as input or use all leads in `LUDBTrainCfg.leads`
)

if LUDBTrainCfg.use_single_lead:
    LUDBTrainCfg.n_leads = 1
else:
    LUDBTrainCfg.n_leads = len(LUDBTrainCfg.leads)

# as for `start_from` and `end_at`, see ref. [1] section 3.1
LUDBTrainCfg.start_from = int(2 * LUDBTrainCfg.fs)
LUDBTrainCfg.end_at = int(2 * LUDBTrainCfg.fs)
LUDBTrainCfg.input_len = int(4 * LUDBTrainCfg.fs)

LUDBTrainCfg.over_sampling = 1

# configs of training epochs, batch, etc.
LUDBTrainCfg.n_epochs = 300
LUDBTrainCfg.batch_size = 32
# LUDBTrainCfg.max_batches = 500500

# configs of optimizers and lr_schedulers
LUDBTrainCfg.optimizer = "adamw_amsgrad"  # "sgd", "adam", "adamw"
LUDBTrainCfg.momentum = 0.949  # default values for corresponding PyTorch optimizers
LUDBTrainCfg.betas = (0.9, 0.999)  # default values for corresponding PyTorch optimizers
LUDBTrainCfg.decay = 1e-2  # default values for corresponding PyTorch optimizers

LUDBTrainCfg.learning_rate = 1e-3  # 1e-4
LUDBTrainCfg.lr = LUDBTrainCfg.learning_rate

LUDBTrainCfg.lr_scheduler = (
    "one_cycle"  # "one_cycle", "plateau", "burn_in", "step", None
)
LUDBTrainCfg.lr_step_size = 50
LUDBTrainCfg.lr_gamma = 0.1
LUDBTrainCfg.max_lr = 2e-3  # for "one_cycle" scheduler, to adjust via expriments

LUDBTrainCfg.burn_in = 400
LUDBTrainCfg.steps = [5000, 10000]

LUDBTrainCfg.early_stopping = CFG()  # early stopping according to challenge metric
LUDBTrainCfg.early_stopping.min_delta = 0.001  # should be non-negative
LUDBTrainCfg.early_stopping.patience = 10

# configs of loss function
LUDBTrainCfg.loss = (
    "FocalLoss"  # "BCEWithLogitsLoss", "AsymmetricLoss", "CrossEntropyLoss"
)
LUDBTrainCfg.loss_kw = CFG()  # "BCEWithLogitsLoss", "AsymmetricLoss"
LUDBTrainCfg.flooding_level = 0.0  # flooding performed if positive

LUDBTrainCfg.log_every = 1
LUDBTrainCfg.monitor = "f1_score"

LUDBTrainCfg.model_name = "unet"



"""
"""

from numbers import Real
from typing import Any, Iterable, Union

import torch

from ..utils.utils_signal_t import normalize as normalize_t

__all__ = [
    "Normalize",
    "MinMaxNormalize",
    "NaiveNormalize",
    "ZScoreNormalize",
]


class Normalize(torch.nn.Module):
    """Normalization preprocessor.

    This preprocessor
    performs z-score normalization on `sig`,
    to make it has fixed mean and standard deviation,
    or performs min-max normalization on `sig`,
    or normalizes `sig` using `mean` and `std`
    :math:`via (sig - mean) / std`. More precisely,

    .. math::

        \\begin{align*}
        \\text{Min-Max normalization:} & \\frac{sig - \\min(sig)}{\\max(sig) - \\min(sig)} \\\\
        \\text{Naive normalization:} & \\frac{sig - m}{s} \\\\
        \\text{Z-score normalization:} & \\left(\\frac{sig - mean(sig)}{std(sig)}\\right) \\cdot s + m
        \\end{align*}

    Parameters
    ----------
    method : {"naive", "min-max", "z-score"}, default "z-score",
        Normalization method, by default "z-score", case-insensitive.
    mean : numbers.Real or array_like, default 0.0
        If `method` is "z-score", then `mean is the mean value
        of the normalized signal,
        or mean values for each lead of the normalized signal.
        If `method` is "naive", then `mean` is the mean value
        to be subtracted from the original signal.
        Useless if `method` is ``"min-max"``.
    std : numbers.Real or array_like, default 1.0
        If `method` is "z-score", then `std` is the standard deviation
        of the normalized signal,
        or standard deviations for each lead of the normalized signal.
        If `method` is "naive", then `std` is the standard deviation
        to be divided from the original signal.
        Useless if `method` is ``"min-max"``.
    per_channel : bool, default False
        Whether to perform the normalization per channel.
    inplace : bool, default True
        Whether to perform the normalization in-place.

    """

    __name__ = "Normalize"

    def __init__(
        self,
        method: str = "z-score",
        mean: Union[Real, Iterable[Real]] = 0.0,
        std: Union[Real, Iterable[Real]] = 1.0,
        per_channel: bool = False,
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.method = method.lower()
        assert self.method in [
            "z-score",
            "naive",
            "min-max",
        ]
        self.mean = mean
        self.std = std
        self.per_channel = per_channel
        self.inplace = inplace

    def forward(self, sig: torch.Tensor) -> torch.Tensor:
        """Apply the preprocessor to the signal tensor.

        Parameters
        ----------
        sig : torch.Tensor
            The input signal tensor,
            of shape ``(batch_size, num_leads, num_samples)``.

        Returns
        -------
        torch.Tensor
            The normalized signal tensor.

        """
        sig = normalize_t(
            sig=sig,
            method=self.method,
            mean=self.mean,
            std=self.std,
            per_channel=self.per_channel,
            inplace=self.inplace,
        )
        return sig


class MinMaxNormalize(Normalize):
    """Min-Max normalization.

    Min-Max normalization is defined as

    .. math::

        \\frac{sig - \\min(sig)}{\\max(sig) - \\min(sig)}

    Parameters
    ----------
    per_channel : bool, default False
        Whether to perform the normalization per channel.
    inplace : bool, default True
        Whether to perform the normalization in-place.

    """

    __name__ = "MinMaxNormalize"

    def __init__(
        self, per_channel: bool = False, inplace: bool = True, **kwargs: Any
    ) -> None:
        super().__init__(
            method="min-max", per_channel=per_channel, inplace=inplace, **kwargs
        )


class NaiveNormalize(Normalize):
    """Naive normalization

    Naive normalization is done via

    .. math::

        \\frac{sig - mean}{std}

    Parameters
    ----------
    mean : numbers.Real or array_like, default 0.0
        Value(s) to be subtracted.
    std : numbers.Real or array_like, default 1.0
        Value(s) to be divided.
    per_channel : bool, default False
        Whether to perform the normalization per channel.
    inplace : bool, default True
        Whether to perform the normalization in-place.

    """

    __name__ = "NaiveNormalize"

    def __init__(
        self,
        mean: Union[Real, Iterable[Real]] = 0.0,
        std: Union[Real, Iterable[Real]] = 1.0,
        per_channel: bool = False,
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__(
            method="naive",
            mean=mean,
            std=std,
            per_channel=per_channel,
            inplace=inplace,
        )


class ZScoreNormalize(Normalize):
    """Z-score normalization.

    Z-score normalization is defined as

    .. math::

        \\left(\\frac{sig - mean(sig)}{std(sig)}\\right) \\cdot s + m

    Parameters
    ----------
    mean : numbers.Real or array_like, default 0.0
        Mean value of the normalized signal,
        or mean values for each lead of the normalized signal.
    std : numbers.Real or array_like, default 1.0
        Standard deviation of the normalized signal,
        or standard deviations for each lead of the normalized signal.
    per_channel : bool, default False
        Whether to perform the normalization per channel.
    inplace : bool, default True
        Whether to perform the normalization in-place.

    """

    __name__ = "ZScoreNormalize"

    def __init__(
        self,
        mean: Union[Real, Iterable[Real]] = 0.0,
        std: Union[Real, Iterable[Real]] = 1.0,
        per_channel: bool = False,
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__(
            method="z-score",
            mean=mean,
            std=std,
            per_channel=per_channel,
            inplace=inplace,
            **kwargs
        )
